from .variant cimport Variant
from cpython cimport array
import array

cdef tuple split(object data, str cigarstring, int target_pos, int string_pos, bint is_for_ref, bint reverse)

cdef tuple locate_indels(str cigarstring, int aln_start_pos)

cdef list get_spliced_subreads(str cigarstring, int read_start_pos, int read_end_pos)

cdef int count_lowqual_non_ref_bases(str read_seq, str ref_seq, array.array quals, list cigar_list, int basequalthresh)



# 786
# Aldy source: test_diplotype_real.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest  # noqa

from .test_diplotype_synthetic import assert_diplotype


def test_basic(real_gene):
    assert_diplotype(real_gene, "*1 / *1", ["1", "1"])


def test_tandem(real_gene):
    assert_diplotype(real_gene, "*2 + *2 / *71", ["2", "2", "71"])
    assert_diplotype(real_gene, "*4.021 + *4C / *41", ["4C", "41", "4.021"])
    assert_diplotype(real_gene, "*4 + *4.019 / *4J", ["4.019", "4", "4J"])
    assert_diplotype(real_gene, "*3 / *68 + *4", ["3", "4", "68"])


def test_36(real_gene):
    assert_diplotype(real_gene, "*36 + *10 / *36 + *41", ["10", "36", "41", "36"])
    assert_diplotype(real_gene, "*1 + *36 / *36 + *10", ["1", "10", "36", "36"])
    assert_diplotype(real_gene, "*10 / *36 + *10", ["10", "36", "10"])
    assert_diplotype(real_gene, "*36 + *10 / *36 + *10", ["10", "36", "36", "10"])
    assert_diplotype(real_gene, "*10 + *10 / *36 + *10", ["10", "36", "10", "10"])


def test_fusion(real_gene):
    assert_diplotype(real_gene, "*2 + *2 / *68 + *4", ["2", "4", "68", "2"])
    assert_diplotype(real_gene, "*1 / *79 + *2", ["1", "2", "79#2"])
    assert_diplotype(real_gene, "*2 / *78 + *2", ["2", "78#2", "2"])
    assert_diplotype(real_gene, "*1 / *78 + *2", ["2", "1", "78#2"])



from .variant import *
from .varaln import *
from .utilities import *
from .consensus import *
from .pileup import *
from .contig import *
from .local_reference import *
from .version import __version__



#cython: profile=False

import re
from cpython cimport array
import array

import numpy as np
from collections import namedtuple
from operator import mul
from functools import reduce
from pysam.libcbcf cimport VariantRecord, VariantRecordFilter, VariantFile

from .variant cimport Variant
from .local_reference cimport UnsplicedLocalReference

cigar_ptrn = re.compile(r"[0-9]+[MIDNSHPX=]")


def most_common(lst):
    alst = list(set(lst))
    alst.sort()
    return max(alst, key=lst.count)


def get_gap_ptrn(read):
    return "".join([c for c in read["cigar_list"] if "D" in c or "I" in c])

def get_gap_ptrn2(read):
    ptrn = ""
    pos = read["aln_start"]
    for cigar in read["cigar_list"]:
        event, event_len = cigar[-1], int(cigar[:-1])
        if event in ("M", "X", "="):
            pos += event_len
        elif event in ["I", "D", "N"]:
            ptrn += "{}@{}".format(cigar, pos-1)
            if event == "D":
                pos += event_len
    return ptrn

def most_common_gap_pattern(targetpileup):
    ptrns = [get_gap_ptrn(read) for read in targetpileup]
    return most_common(ptrns)

def most_common_gap_ptrn(targetpileup):
    ptrns = [get_gap_ptrn2(read) for read in targetpileup]
    return most_common(ptrns)

cpdef list to_flat_list(list lst_of_lst):
    cdef list lst
    return [i for lst in lst_of_lst for i in lst]


cpdef list to_flat_vcf_records(VariantRecord record):

    cdef str alt

    VcfRec = namedtuple(
        "VcfRec", "chrom pos id ref alt qual filter info format samples orig"
    )

    if not record.alts:
        return []

    flat_record = [
        VcfRec(
            chrom=record.chrom,
            pos=record.pos,
            id=record.id,
            ref=record.ref,
            alt=alt,
            qual=record.qual,
            filter=record.filter,
            info=record.info,
            format=record.format,
            samples=record.samples,
            orig=record,
        )
        for alt in record.alts
    ]

    return flat_record


cpdef dict to_dict(object record):
    cdef str k

    d = {}
    for k, v in record.items():
        if isinstance(v, tuple):
            d[k] = ",".join([str(i) for i in v])
        else:
            d[k] = v

    if d:
        return d


cpdef bint match_indels(Variant query, Variant subject, str matchby, bint indel_only):
    if matchby != "normalization" and indel_only and not query.is_indel:
        return False

    if matchby == "normalization":
        return query == subject

    elif matchby == "locus":
        if query.chrom != subject.chrom:
            return False

        query.normalize(inplace=True)
        subject.normalize(inplace=True)

        return query.pos == subject.pos

    elif matchby == "exact":
        return (
            (query.chrom == subject.chrom)
            and (query.pos == subject.pos)
            and (query.ref == subject.ref)
            and (query.alt == subject.alt)
        )


cpdef double linguistic_complexity(str seq):
    cdef int i, j, n
    n = len(seq)
    if n <= 1:
        return float(n)
    else:
        usage = []
        for i in range(1, n):
            i_mer = [seq[j : j + i] for j in range(n - i + 1)]
            usage.append(len(set(i_mer)) / min(4 ** i, n - i + 1))

        return reduce(mul, usage)


cpdef double low_qual_fraction(list pileup):
    cdef dict read
    cdef int pileup_vol = 1
    cdef int low_qual_vol = 0

    for read in pileup:
        pileup_vol += len(read["read_seq"])
        low_qual_vol += read["low_qual_base_num"]

    return low_qual_vol / pileup_vol


def to_minimal_repeat_unit(seq):
    """Find repeat unit in indel sequence
    """
    mid = int(len(seq) / 2)
    min_unit = seq

    j = 1
    found = False

    while j <= mid and not found:
        tandems = [seq[i : i + j] for i in range(0, len(seq), j)]
        if len(set(tandems)) == 1:
            found = True
            min_unit = list(set(tandems))[0]
        j += 1

    return min_unit


def repeat_counter(query_seq, flank_seq):
    """
    """
    qlen, flen = len(query_seq), len(flank_seq)
    count = 0

    if flen < qlen:
        return count

    for i in range(0, flen, qlen):
        if flank_seq[i  : i + qlen] == query_seq:
            count += 1
        else:
            break

    return count


cdef int count_lowqual_non_ref_bases(
    str read_seq,
    str ref_seq,
    array.array quals,
    list cigar_list,
    int basequalthresh,
):
    cdef int i = 0, j = 0, k = 0, cnt = 0, event_len = 0
    cdef str event = "", cigar = "" ,

    for cigar in cigar_list:
        event, event_len = cigar[-1], int(cigar[:-1])

        if event in ("M", "=", "X"):
            while k < event_len:
                if read_seq[i] != ref_seq[j] and quals[i] < basequalthresh:
                    cnt += 1
                i += 1
                j += 1
                k += 1
            k = 0
        elif event in ("I", "S"):
            while k < event_len:
                if quals[i] < basequalthresh:
                    cnt += 1
                i += 1
                k += 1
            k = 0
        elif event == "D":
            j += event_len

    return cnt


cpdef list get_mapped_subreads(str cigarstring, int aln_start_pos, int aln_end_pos):

    cdef int event_len, current_pos
    cdef str cigar, event
    cdef list cigar_lst = cigar_ptrn.findall(cigarstring)
    cdef list res = []

    current_pos = aln_start_pos
    for cigar in cigar_lst:
        event, event_len = cigar[-1], int(cigar[:-1])

        if event in ("M", "X", "="):
            res.append((current_pos, (current_pos + event_len - 1)))
            current_pos += event_len
        elif event in ("I", "S", "H", "P"):
            pass
        else:
            current_pos += event_len

    return res


cdef list get_spliced_subreads(str cigarstring, int read_start_pos, int read_end_pos):

    cdef int i = 0
    cdef int event_len
    cdef str cigar, event, prev_event
    cdef list cigar_lst = cigar_ptrn.findall(cigarstring)
    cdef list pos_lst = [read_start_pos]
    cdef list res = []

    if not "N" in cigarstring:
        return [(read_start_pos, read_end_pos)]

    prev_event = "A"
    for cigar in cigar_lst:
        event, event_len = cigar[-1], int(cigar[:-1])

        if event == "N":
            pos_lst.append(read_start_pos - 1)
        elif prev_event == "N":
            pos_lst.append(read_start_pos)

        if event in ("I", "H", "P"):
            pass
        else:
            read_start_pos += event_len

        prev_event = event

    if prev_event != "N":
        pos_lst.append(read_end_pos)

    while i < len(pos_lst):
        res.append(pos_lst[i : i+2])
        i += 2

    return res


cpdef int get_end_pos(int read_start_pos, str lt_flank, str cigarstring):
    read_start_pos -= 1

    cdef list cigar_lst = cigar_ptrn.findall(cigarstring)
    cdef str cigar, event
    cdef int event_len, i = 0, flank_len = len(lt_flank)

    while flank_len > 0:
        cigar = cigar_lst[i]
        event, event_len = cigar[-1], int(cigar[:-1])

        if event == "D" or event == "N":
            read_start_pos += event_len
        elif event == "I":
            flank_len -= event_len
        elif event == "H" or event == "P":
            pass
        else:
            flank_len -= event_len
            read_start_pos += event_len

        i += 1

    return read_start_pos + flank_len


cdef tuple locate_indels(str cigarstring, int aln_start_pos):
    aln_start_pos -= 1

    cdef list cigar_lst = cigar_ptrn.findall(cigarstring)
    cdef str cigar, event
    cdef int event_len

    cdef list ins = []
    cdef list dels = []
    for cigar in cigar_lst:
        event, event_len = cigar[-1], int(cigar[:-1])
        if event == "I":
            ins.append((aln_start_pos, event_len))
        elif event == "D":
            dels.append((aln_start_pos, event_len))
            aln_start_pos += event_len
        elif event == "H" or event == "P":
            pass
        else:
            aln_start_pos += event_len

    return ins, dels


cpdef tuple split_cigar(str cigarstring, int target_pos, int start):

    cdef list cigar_lst = cigar_ptrn.findall(cigarstring)
    cdef str cigar, event
    cdef int event_len, move, diff
    cdef list lt_lst = []
    cdef list rt_lst = cigar_lst

    start -= 1
    for cigar in cigar_lst:
        event, event_len = cigar[-1], int(cigar[:-1])

        move = 0 if event in ("I", "H", "P") else event_len
        start += move
        rt_lst = rt_lst[1 :]

        if target_pos <= start:
            diff = start - target_pos
            lt_cigar = str(event_len - diff) + event
            lt_lst.append(lt_cigar)

            if diff:
                    rt_lst = [str(diff) + event] + rt_lst

            return lt_lst, rt_lst
        else:
            lt_lst.append(cigar)


def merge_consecutive_gaps(cigar_lst):

    merged_lst = []
    while cigar_lst:
        c = cigar_lst[0]
        cigar_lst = cigar_lst[1:]

        if "I" in c or "D" in c:
            i = 0
            is_gap = True
            while i < len(cigar_lst) and is_gap:
                tmp = cigar_lst[i]
                is_gap = True if "I" in tmp or "D" in tmp else False
                i += 1

            if i - 1:
                c += "".join(cigar_lst[: i - 1])
                cigar_lst = cigar_lst[i - 1 :]

        merged_lst.append(c)

    return merged_lst


def make_insertion_first(cigarstring):

    cigar_lst = cigar_ptrn.findall(cigarstring)

    merged_cigar_lst = merge_consecutive_gaps(cigar_lst)
    new_cigar = []
    for c in merged_cigar_lst:
        if "I" in c and "D" in c:
            c_lst = cigar_ptrn.findall(c)
            if "D" in c_lst[0]:
                swapped = c_lst[::-1]
                new_cigar.append("".join(swapped))
            else:
                new_cigar.append("".join(c_lst))
        else:
            new_cigar.append(c)

    return "".join(new_cigar)


def relative_aln_pos(ref_seq, cigar_lst, aln_start, target_pos, include_clip=False):

    current_pos = aln_start - 1
    ref_seq_pos = 0
    for cigar in cigar_lst:
        event, event_len = cigar[-1], int(cigar[:-1])

        event = "M" if include_clip and event == "S" else event

        if event == "M" or event == "D":
            current_pos += event_len
            ref_seq_pos += event_len
        elif event in ("I", "H", "P"):
            pass
        else:
            current_pos += event_len

        if current_pos >= target_pos:
            break

    ref_seq_pos += (target_pos - current_pos)

    return ref_seq_pos / len(ref_seq)


cdef tuple split(
    object data,
    str cigarstring,
    int target_pos,
    int string_pos,
    bint is_for_ref,
    bint reverse,
):

    cdef list cigar_lst = cigar_ptrn.findall(cigarstring)
    cdef int _size = len(cigar_lst)

    cdef str cigar, event
    cdef int event_len, d_move, g_move

    cdef double [:] data_moves = np.zeros((_size,))
    cdef double [:] genome_moves = np.zeros((_size,))

    cdef int i = 0, j = 0

    for cigar in cigar_lst:
        event, event_len = cigar[-1], int(cigar[:-1])

        if event == "N":
            d_move = 0
            g_move = event_len
        elif event == "I":
            g_move = 0
            d_move = 0 if is_for_ref else event_len
        elif event == "D":
            g_move = event_len
            d_move = event_len if is_for_ref else 0
        elif event == "H" or event == "P":
            d_move = 0
            g_move = 0
        else:
            g_move, d_move = event_len, event_len

        data_moves[i] = d_move
        genome_moves[i] = g_move
        i += 1

    if reverse:
        string_pos += 1
        data = data[::-1]
        data_moves = data_moves[::-1]
        genome_moves = genome_moves[::-1]
    else:
        string_pos -= 1

    for d_move, g_move in zip(data_moves, genome_moves):
        if reverse:
            if target_pos < string_pos:
                string_pos -= g_move
            else:
                break
        else:
            if string_pos < target_pos:
                string_pos += g_move
            else:
                break
        j += d_move

    diff = string_pos - (target_pos + 1)if reverse else target_pos - string_pos
    if reverse:
        lt = data[j + diff :]
        lt = lt[::-1]
        rt = data[: j + diff]
        rt = rt[::-1]
    else:
        lt = data[: j + diff]
        rt = data[j + diff :]

    return lt, rt


cpdef tuple get_local_reference(
    Variant target,
    list pileup,
    int window,
    UnsplicedLocalReference unspl_loc_ref,
    bint unspliced=False,
    bint splice_pattern_only=False,
):

    cdef str span
    cdef tuple ptrn

    chrom, pos, reference = target.chrom, target.pos, target.reference

    if unspliced:
        splice_patterns = None
    else:
        splice_patterns = [read["splice_pattern"] for read in pileup if read["splice_pattern"] != ("", "")]

    ref_len = reference.get_reference_length(chrom)

    cdef list spl_ptrn = []

    if splice_patterns:
        lt_patterns = [ptrn[0] for ptrn in splice_patterns if ptrn[0]]
        if lt_patterns:
            lt_pattern = most_common(lt_patterns)
            lt_spl_pos = []
            for span in lt_pattern.split(":"):
                lt_spl_pos += [int(i) for i in span.split("-")]
        else:
            lt_spl_pos = []

        rt_patterns = [ptrn[1] for ptrn in splice_patterns if ptrn[1]]
        if rt_patterns:
            rt_pattern = most_common(rt_patterns)
            rt_spl_pos = []
            for span in rt_pattern.split(":"):
                rt_spl_pos += [int(i) for i in span.split("-")]
        else:
            rt_spl_pos = []

        spl_pos = lt_spl_pos + rt_spl_pos
        last_idx = len(spl_pos) - 1

        left_len = 0
        first_pass = False
        local_reference = ""
        for i, x in enumerate(spl_pos):
            if i == 0:
                lt_end = max(0, x - window * 2)
                local_reference += reference.fetch(chrom, lt_end, x - 1)
                rt_end = x - 1
                if x + 1 < rt_end:
                    spl_ptrn.append((x + 1, rt_end))
                else:
                    spl_ptrn.append((lt_end, rt_end))
            elif i % 2 == 1 and i != last_idx:
                local_reference += reference.fetch(chrom, x, spl_pos[i+1] - 1)
                rt_end = spl_pos[i+1] - 1
                spl_ptrn.append((x + 1, rt_end))
            elif i % 2 == 0:
                pass
            elif i == last_idx:
                rt_end = min(x + window * 2, ref_len)
                local_reference += reference.fetch(chrom, x, rt_end)
                spl_ptrn.append((x + 1, rt_end))

            if pos <= rt_end and not first_pass:
                left_len = len(local_reference) - (rt_end - pos)
                first_pass = True

    else:
        local_reference = unspl_loc_ref.fetch_ref_seq(pos, window)
        #left_len = unspl_loc_ref.left_len
        #local_reference = reference.fetch(chrom, max(0, pos - window * 3), min(pos + window * 3, ref_len))
        left_len = pos - max(0, pos - window * 3)

    if splice_pattern_only:
        return tuple(spl_ptrn)

    return local_reference, left_len



# 786
# Aldy source: solutions.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Dict
from dataclasses import dataclass, field
from natsort import natsorted
from collections import Counter, defaultdict

from .gene import Gene, Mutation
from .profile import Profile


@dataclass
class CNSolution:
    """Valid copy-number configuration assignment."""

    gene: Gene
    """Gene instance."""
    score: float
    """ILP model objective score (0 for user-provided solutions)."""
    solution: Dict[str, int]
    """
    Copy-number configurations mapped to the corresponding copy number
    (e.g. `{1: 2}` means that there are two copies of `*1` configuration).
    """
    region_cn: List[Dict[str, int]]
    """Gene region copy numbers inferred by this solution."""

    def __init__(self, gene: Gene, score: float, solution: List[str]):
        self.region_cn = [
            {r: 0 for r in gene.regions[0]} for _ in gene.cn_configs["1"].cn
        ]
        for conf in solution:
            for gi, g in enumerate(gene.cn_configs[conf].cn):
                for r in g:
                    self.region_cn[gi][r] += g[r]
        self.gene = gene
        self.score = score
        self.solution = Counter(solution)

    def __hash__(self):
        return str(self).__hash__()

    def position_cn(self, pos: int) -> float:
        """:returns: Copy number at the given location."""
        r = self.gene.region_at(pos)
        return self.region_cn[r[0]][r[1]] if r else 0

    def _solution_nice(self):
        return ",".join(f"{v}x*{k}" for k, v in natsorted(self.solution.items()))

    def __str__(self):
        return "CNSol[{:.2f}; sol=({}); cn={}]".format(
            self.score,
            self._solution_nice(),
            "|".join(
                "".join(str(self.region_cn[g][r]) for r in self.gene.regions[0])
                for g, _ in enumerate(self.region_cn)
            ),
        )

    def max_cn(self):
        """Maximum copy-number in the solution."""
        return sum(self.solution.values())


@dataclass
class SolvedAllele:
    """Valid star-allele assignment."""

    gene: Gene
    """Gene instance."""
    major: str
    """Major star-allele identifier."""
    minor: str = ""
    """Minor star-allele identifier. Empty string when not assigned."""
    added: List[Mutation] = field(default_factory=list)
    """Mutations that are added to the star-allele (from the allele database)."""
    missing: List[Mutation] = field(default_factory=list)
    """
    Mutations that are omitted from the star-allele (but present in the star-allele
    definition).
    """

    def mutations(self):
        """Set of allele mutations."""
        m = self.gene.alleles[self.major].func_muts
        if self.minor:
            m |= self.gene.alleles[self.major].minors[self.minor].neutral_muts
        m |= set(self.added)
        m -= set(self.missing)
        return m

    def major_repr(self):
        """Pretty-formats major star-allele name."""
        extra = "".join(
            " +" + self.gene.get_rsid(m)
            for m in sorted(m for m in self.added if self.gene.is_functional(m, False))
        )
        s = f"{self.major}{extra}"
        return f"*({s})" if extra else f"*{s}"

    def __str__(self):
        """Pretty-formats minor star-allele name."""
        extra = "".join(" +" + self.gene.get_rsid(m) for m in sorted(self.added))
        miss = "".join(" -" + self.gene.get_rsid(m) for m in sorted(self.missing))
        s = f"{self.minor if self.minor else self.major}{extra}{miss}"
        return f"*({s})" if extra or miss else f"*{s}"

    def __hash__(self):
        return hash(
            (self.gene.name, self.major, self.minor, *self.added, *self.missing)
        )


@dataclass
class MajorSolution:
    """Valid major star-allele assignment."""

    score: float
    """ILP model objective score."""
    solution: Dict[SolvedAllele, int]
    """
    Major star-alleles and the corresponding copy numbers
    (e.g. `{1: 2}` means that we have two copies of `*1`).
    """
    cn_solution: CNSolution
    """Copy-number solution that was used to assign major star-alleles."""
    added: List[Mutation]
    """
    List of added mutations. Will be assigned to :py:class:`SolvedAllele` in the
    minor star-allele calling step if phasing is enabled.
    """

    def _solution_nice(self):
        x = ", ".join(
            f"{v}x{s}"
            for s, v in natsorted(self.solution.items(), key=lambda x: x[0].major)
        )
        y = ", ".join(self.cn_solution.gene.get_rsid(m) for m in sorted(self.added))
        return " & ".join([x, y] if y else [x])

    def __str__(self):
        return (
            f"MajorSol[{self.score:.2f}; "
            + f"sol=({self._solution_nice()}); "
            + f"cn={self.cn_solution}"
        )

    def __hash__(self):
        return str(self).__hash__()


@dataclass
class MinorSolution:
    """Valid minor star-allele assignment."""

    score: float
    """ILP model objective score."""
    solution: List[SolvedAllele]
    """
    List of solved minor star-alleles.
    Modifications to minor alleles are represented in :py:class:`SolvedAllele`.
    """
    major_solution: MajorSolution
    """Major star-allele solution used for calculating minor star-allele assignments."""
    profile: Profile = None
    """Profile."""

    def _solution_nice(self):
        return ", ".join(
            str(s) for s in natsorted(self.solution, key=lambda x: x.minor)
        )

    def __str__(self):
        return (
            f"MinorSol[{self.score:.2f}; "
            + f"sol=({self._solution_nice()}); "
            + f"major={self.major_solution}"
        )

    def get_diplotype(self):
        try:
            return self.diplotype
        except AttributeError:
            self.diplotype = ([], [])
            return self.diplotype

    def set_diplotype(self, d):
        self.diplotype = d

    def get_major_name(self, i):
        gene = self.major_solution.cn_solution.gene
        if i == -1:
            return gene.deletion_allele()
        n = str(self.solution[i].major).split("#")[0:1]

        def get_nice_snp_name(m):
            rsid = gene.get_rsid(m)
            if not self.profile or not self.profile.display_format:
                return rsid
            effect = gene.mutations[m][0]
            al = [
                an
                for an, a in gene.alleles.items()
                if m in a.func_muts and len(a.func_muts) == 1
            ]
            if al:
                return f"{al[0]}.{rsid}"
            elif effect:
                return f"{effect}.{rsid}"
            return rsid

        for m in sorted(self.solution[i].added):
            if gene.is_functional(m, infer=False):
                n.append(get_nice_snp_name(m))
        if not self.profile or not self.profile.display_format:
            return "+".join(n)
        elif len(n) == 1:
            return n[0]
        else:
            return "(" + " ^".join(n) + ")"

    def get_minor_name(self, i, legacy=False):
        gene = self.major_solution.cn_solution.gene
        if i == -1:
            return gene.deletion_allele()

        m = gene.alleles[self.solution[i].major]
        t = [mi for mi in m.minors if mi == self.solution[i].minor]
        assert len(t) == 1
        n = [m.minors[t[0]].alt_name or t[0] if legacy else t[0]]
        for m in sorted(self.solution[i].added):
            n.append("+" + gene.get_rsid(m))
        for m in sorted(self.solution[i].missing):
            n.append("-" + gene.get_rsid(m))
        return " ".join(n)

    def get_major_diplotype(self):
        return " / ".join(
            " + ".join(f"*{self.get_major_name(i)}" for i in d)
            for d in self.diplotype
            if d
        )

    def get_minor_diplotype(self, legacy=False):
        return " / ".join(
            " + ".join(f"[*{self.get_minor_name(i, legacy)}]" for i in d)
            for d in self.diplotype
            if d
        )

    def get_mutation_coverages(self, coverage):
        muts = defaultdict(int)
        for sa in self.solution:
            for m in (
                sa.gene.alleles[sa.major].func_muts
                | sa.gene.alleles[sa.major].minors[sa.minor].neutral_muts
            ):
                if m not in sa.missing:
                    muts[m] += 1
            for m in sa.added:
                muts[m] += 1
        covs = {}
        for m in muts:
            covs[m] = coverage.single_copy(m.pos, self.major_solution.cn_solution)
            covs[m] = coverage[m] / covs[m] if covs[m] > 0 else 0
        return [(m, muts[m], covs[m]) for m in muts]



#!/usr/bin/env python3
import re
import numpy as np
import traceback
from collections import OrderedDict

from .utilities import (
    most_common,
    get_gap_ptrn2,
    most_common_gap_ptrn,
    to_flat_list,
    merge_consecutive_gaps,
)

cigar_ptrn = re.compile(r"[0-9]+[MIDNSHPX=]")


def make_consensus(target, targetpileup, basequalthresh):
    expected_ptrn = most_common_gap_ptrn(targetpileup)
    targetpileup = [
        read for read in targetpileup if expected_ptrn == get_gap_ptrn2(read)
    ]

    target_pos, target_type, target_len = (
        target.pos,
        target.variant_type,
        len(target.indel_seq),
    )

    lt_indexed, rt_indexed, contributing_reads = [], [], []
    for read in targetpileup:
        target_pos = target_pos if read.get("target_right_shifted", 0) else target_pos

        try:
            lt = index_bases(
                read["read_start"],
                target_pos,
                target_type,
                target_len,
                read["lt_cigar"],
                read["lt_flank"],
                read["lt_ref"],
                read["lt_qual"],
            )
            rt = index_bases(
                read["read_start"],
                max(k for k, v in lt.items()),
                target_type,
                target_len,
                read["rt_cigar"],
                read["rt_flank"],
                read["rt_ref"],
                read["rt_qual"],
                left_padding=(
                    read["lt_ref"][-1],
                    read["lt_flank"][-1],
                    read["lt_qual"][-1],
                    target.indel_seq,
                ),
                left_flank=read["lt_flank"],
                left=False,
            )
            lt_indexed.append(lt)
            rt_indexed.append(rt)
            contributing_reads.append(read)

        except:
            # this is for debugging
            # traceback.print_exc()
            pass

    if lt_indexed and rt_indexed:
        lt_consensus = consensus_data(lt_indexed, True, basequalthresh)
        rt_consensus = consensus_data(rt_indexed, False, basequalthresh)

        return lt_consensus, rt_consensus, contributing_reads


def index_bases(
    read_pos,
    target_pos,
    target_type,
    target_len,
    cigar,
    flank,
    ref,
    qual,
    left_padding=None,
    left_flank=None,
    left=True,
):
    indexedbases = {}

    orig = cigar
    cigar = merge_consecutive_gaps(cigar)

    if left:
        current_pos = read_pos
    else:
        current_pos = target_pos
        ref_padding, flank_padding, qual_padding, target_seq = (
            left_padding[0],
            left_padding[1],
            left_padding[2],
            left_padding[3],
        )

        first_cigar = cigar_ptrn.findall(cigar[0])
        if len(first_cigar) == 1 and target_type in first_cigar[0]:
            if target_type == "I":
                indexedbases[current_pos] = (
                    ref_padding,
                    flank_padding + target_seq,
                    qual_padding,
                )
                current_pos += 1
            else:
                indexedbases[current_pos] = (
                    ref_padding + target_seq,
                    flank_padding,
                    qual_padding,
                )
                current_pos += target_len + 1
        elif len(first_cigar) == 2:
            # ins_first = "I" in first_cigar[0]

            del_len = sum(int(c[:-1]) for c in first_cigar if c[-1] == "D")
            ins_len = sum(int(c[:-1]) for c in first_cigar if c[-1] == "I")

            if target_type == "I":
                indexedbases[current_pos] = (
                    ref_padding + ref[:del_len],
                    flank_padding + target_seq,
                    qual_padding,
                )
                ref = ref[del_len:]
                current_pos += del_len + 1
            elif target_type == "D":
                indexedbases[current_pos] = (
                    ref_padding + target_seq,
                    left_flank[-(ins_len + 1) :],
                    qual_padding,
                )

            current_pos += del_len + 1

        elif len(first_cigar) == 1 and "M" in first_cigar[0]:
            map_len = int(first_cigar[0][:-1])
            for i in range(map_len):
                indexedbases[current_pos] = (ref[0], flank[0], qual[0])
                ref = ref[1:]
                flank = flank[1:]
                qual = qual[1:]
                current_pos += 1

        cigar = cigar[1:]

    for c in cigar:
        # complex pattrn (I and D merged)
        if "I" in c and "D" in c:
            tmp = cigar_ptrn.findall(c)
            ins_len = sum(int(i[:-1]) for i in tmp if i[-1] == "I")
            del_len = sum(int(i[:-1]) for i in tmp if i[-1] == "D")

            padding_ref, padding_qual = (
                indexedbases[current_pos - 1][0],
                indexedbases[current_pos - 1][2],
            )

            ins_seq, flank, ins_qual, qual = (
                flank[:ins_len],
                flank[ins_len:],
                qual[:ins_len],
                qual[ins_len:],
            )

            del_seq, ref = ref[:del_len], ref[del_len:]

            indexedbases[current_pos - 1] = (
                padding_ref + del_seq,
                padding_ref + ins_seq,
                np.median([padding_qual] + list(ins_qual)),
            )

            current_pos += del_len
        else:
            event, event_len = c[-1], int(c[:-1])

            if event in ("M", "S", "X", "="):
                for i in range(event_len):
                    if ref and event != "S":
                        indexedbases[current_pos] = (ref[0], flank[0], qual[0])
                        ref = ref[1:]
                    else:
                        indexedbases[current_pos] = ("", flank[0], qual[0])

                    flank = flank[1:]
                    qual = qual[1:]
                    current_pos += 1

            elif event == "I":
                padding_ref, padding_qual = (
                    indexedbases[current_pos - 1][0],
                    indexedbases[current_pos - 1][2],
                )
                ins_seq, flank, ins_qual, qual = (
                    flank[:event_len],
                    flank[event_len:],
                    qual[:event_len],
                    qual[event_len:],
                )
                indexedbases[current_pos - 1] = (
                    padding_ref,
                    padding_ref + ins_seq,
                    np.median([padding_qual] + list(ins_qual)),
                )

            elif event == "D":
                padding_ref, padding_qual = (
                    indexedbases[current_pos - 1][0],
                    indexedbases[current_pos - 1][2],
                )
                del_seq, ref = ref[:event_len], ref[event_len:]
                indexedbases[current_pos - 1] = (
                    padding_ref + del_seq,
                    padding_ref,
                    padding_qual,
                )

                current_pos += event_len

            elif event == "N":
                current_pos += event_len

    return indexedbases


def consensus_data(indexedbases_list, left, basequalthresh):
    consensus_index = OrderedDict()

    skip_loci = []
    for locus in locus_list(indexedbases_list, left):
        ref, consensus_base, consensus_score, coverage = get_consensus_base(
            indexedbases_list, locus, basequalthresh
        )

        if len(ref) > len(consensus_base) and "N" not in consensus_base:
            del_len = len(ref) - len(consensus_base)
            skip_loci += [locus + i for i in range(1, del_len + 1)]

        consensus_index[locus] = (ref, consensus_base, consensus_score, coverage)

    for locus in skip_loci:
        if locus in consensus_index:
            del consensus_index[locus]

    conseq, refseq = "", ""
    scores = []
    coverages = []
    prev_ref = ""
    prev_locus = -1
    ref_end = -1
    for locus, data in consensus_index.items():
        ref, consensus_base, consensus_score, coverage = (
            data[0],
            data[1],
            data[2],
            data[3],
        )

        if left and len(ref) != len(consensus_base):
            ref = ref[::-1]
            consensus_base = consensus_base[::-1]

        refseq += ref
        conseq += consensus_base
        scores += [consensus_score] * len(consensus_base)
        coverages += [coverage] * len(consensus_base)

        if prev_ref and not ref:
            ref_end = prev_locus

        prev_locus = locus
        prev_ref = ref

    if left:
        conseq = conseq[::-1]
        refseq = refseq[::-1]
        scores = scores[::-1]
        coverages = coverages[::-1]

    return consensus_index, ref_end, refseq, conseq, scores, coverages


def locus_list(dict_list, left):
    loci = to_flat_list([[*d] for d in dict_list])
    loci = list(set(loci))
    loci.sort(reverse=left)
    return loci


def get_consensus_base(indexedbases_list, locus, basequalthresh):
    refs = [
        indexedbases[locus][0].upper()
        for indexedbases in indexedbases_list
        if indexedbases.get(locus, False)
    ]
    bases = [
        indexedbases[locus][1]
        for indexedbases in indexedbases_list
        if indexedbases.get(locus, False) and indexedbases[locus][1] != "N"
    ]
    quals = [
        indexedbases[locus][2]
        for indexedbases in indexedbases_list
        if indexedbases.get(locus, False)
    ]

    if not bases:
        ref = most_common(refs) if refs else ""
        return ref, "N", 0.0, 0

    hq_bases = [base for base, qual in zip(bases, quals) if qual >= basequalthresh]

    refs = refs + [""] * (len(bases) - len(refs))

    pairs = [(ref, base) for ref, base in zip(refs, bases)]

    consensus_pair = most_common(pairs)
    consensus_score = pairs.count(consensus_pair) / len(pairs)

    ref, consensus_base = consensus_pair[0], consensus_pair[1]

    if ref:
        if ref != consensus_base:
            if not consensus_base in hq_bases:
                consensus_base = "N"
                consensus_score = 0.0
    else:
        if hq_bases:
            consensus_base = most_common(hq_bases)
            consensus_score = bases.count(consensus_base) / len(bases)
        else:
            consensus_base = "N"
            consensus_score = 0.0

    return ref, consensus_base, consensus_score, len(pairs)


def consensus_refseq(refseq_lst, left=False):
    if left:
        refseq_lst = [seq[::-1].upper() for seq in refseq_lst]
    else:
        refseq_lst = [seq.upper() for seq in refseq_lst]

    consensus_seq = ""
    consensus_rates = []
    for i in range(len(max(refseq_lst, key=len))):
        ith_chars = [ith_char(seq, i) for seq in refseq_lst if ith_char(seq, i)]
        cosensus_base = most_common(ith_chars)

        if cosensus_base == "N":
            consensus_rate = 0.0
        else:
            consensus_rate = ith_chars.count(cosensus_base) / len(ith_chars)

        consensus_seq += cosensus_base
        consensus_rates.append(consensus_rate)

    if left:
        consensus_seq = consensus_seq[::-1]
        consensus_rates = consensus_rates[::-1]

    return consensus_seq, consensus_rates


def ith_char(seq, i):
    try:
        return seq[i]
    except:
        return None


def is_compatible(query, subject, indel_type, partial_match=True):
    """Check if query indel looks same as subject indel

    Args:
        query (dict): dictized read
        subject (dict): indel template
        indel_type (str): "I" for ins "D" for del
        patial_match (bool): True to allow partial match for longer insertions
    Returns:
        True/False (bool): True if query indel is found the same as subject indel
    """
    query_lt_flank, query_indel, query_del, query_rt_flank = (
        query["lt_flank"],
        query["indel_seq"],
        query.get("del_seq", ""),
        query["rt_flank"],
    )

    query_indel_seq = query_indel if query_indel else query_del

    # left-align check (if not, an alternative alignment of something else)
    if query_indel_seq and query_lt_flank and query_lt_flank[-1] == query_indel_seq[-1]:
        return False

    (
        subject_lt_flank,
        subject_lt_scores,
        subject_indel,
        subject_rt_flank,
        subject_rt_scores,
    ) = (
        subject.lt_target_block_consensus_seq,
        subject.lt_target_block_consensus_scores,
        subject.indel_seq,
        subject.rt_target_block_consensus_seq,
        subject.rt_target_block_consensus_scores,
    )

    # check flanking sequences for similarity
    lt_len = min(len(query_lt_flank), len(subject_lt_flank))
    rt_len = min(len(query_rt_flank), len(subject_rt_flank))

    if lt_len > 0:
        lt_query, lt_subject, lt_scores = (
            query_lt_flank[-lt_len:],
            subject_lt_flank[-lt_len:],
            subject_lt_scores[-lt_len:],
        )
    else:
        lt_query, lt_subject, lt_scores = "", "", [0]

    rt_query, rt_subject, rt_scores = (
        query_rt_flank[:rt_len],
        subject_rt_flank[:rt_len],
        subject_rt_scores[:rt_len],
    )

    if lt_query and not is_almost_same(
        lt_query[::-1], lt_subject[::-1], lt_scores[::-1]
    ):
        return False

    if rt_query and not is_almost_same(rt_query, rt_subject, rt_scores):
        return False

    # repeat boundary check (assuming left-aligned bam)
    rt_check = contains_repeat_end(subject_indel, rt_query, subject_rt_flank)
    if not rt_check:
        return False

    # check inserted/deleted sequences for similarity
    if query_indel and indel_type == "I":
        subject_len = len(subject_indel)
        query_len = len(query_indel)

        if subject_len < query_len:
            return False
        elif subject_indel == query_indel:
            return True
        elif 4 <= subject_len <= 6 and partial_match:
            return identical_for_end_n_bases(query_indel, subject_indel, 3)
        elif 7 <= subject_len <= 8 and partial_match:
            return identical_for_end_n_bases(query_indel, subject_indel, 4)
        elif 9 <= subject_len <= 10 and partial_match:
            return identical_for_end_n_bases(query_indel, subject_indel, 5)
        elif 11 <= subject_len and partial_match:
            return identical_for_end_n_bases(query_indel, subject_indel, 6)
        else:
            return False
    elif not query_indel and indel_type == "D":
        return True
    else:
        return False


def contains_repeat_end(indel_seq, query_flank, subject_flank):
    """Check if repeat boundary is contained"""
    tmp = subject_flank.replace(indel_seq, "")

    if tmp:
        repeat_end = tmp[0]
    else:
        return False

    if repeat_end == "N":
        return False

    tmp2 = query_flank.replace(indel_seq, "")
    if tmp2:
        repeat_end2 = tmp2[0]
        return repeat_end == repeat_end2
    else:
        return False


def identical_for_end_n_bases(query_str, subject_str, n):
    return (query_str[:n] == subject_str[:n]) or (query_str[-n:] == subject_str[-n:])


def is_almost_same(
    query_seq,
    subject_seq,
    consensus_score,
    consensus_lim=0.7,
    len_lim=2,
    mismatch_lim=2,
):
    """Check if query and subject sequences are same for high consensus bases

    Args:
        query_seq (str): query flanking seq
        subject_seq (str): indel template flanking seq constructed by consensus
        consensus_score (list): consensus score for template flanking seq
        consensus_lim (float): threshold to define "high consensus"
        len_lim (int): seq shorter than len_lim may not contain any mismatch
        mismatch_lim (int): allow (mismatch_lim)x mismatches for seq longer than len_lim
    Returns:
        True/False (bool): True if query and subject look same
    """
    seq_len = len(query_seq)
    if seq_len > 0 and query_seq[0] != subject_seq[0]:
        return False

    # mismatches for high-consensus bases
    mismatches = [
        (query_seq[i] != subject_seq[i] and consensus_score[i] > consensus_lim)
        for i in range(seq_len)
    ]

    if seq_len < len_lim:
        return sum(mismatches) == 0
    else:
        near_mismatches = mismatches[:len_lim]
        mid_mismatches = mismatches[len_lim : (10 * len_lim)]
        far_mismatches = mismatches[(10 * len_lim) : (30 * len_lim)]
        mismatch_score = (
            sum(near_mismatches) * 2 + sum(mid_mismatches) + sum(far_mismatches) * 0.5
        )
        return mismatch_score < mismatch_lim



from .variant cimport Variant
from .local_reference cimport UnsplicedLocalReference
from pysam.libcalignmentfile cimport AlignmentFile

cdef tuple make_pileup(
    Variant target,
    AlignmentFile bam,
    UnsplicedLocalReference unspl_loc_ref,
    bint exclude_duplicates,
    int window,
    int downsamplethresh,
    int basequalthresh
)



# 786
# Aldy source: sam.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Tuple, Dict, List, Optional
from collections import defaultdict, Counter
from statistics import mean
import pysam
import os
import os.path
import gzip
import tarfile
import pickle
import tempfile

from .indelpost import Variant, VariantAlignment
from .common import log, GRange, AldyException, script_path, Timing, chr_prefix
from .gene import Gene, CNConfigType
from .coverage import Coverage
from .profile import Profile


class Sample:
    """Parse read alignments in a SAM/BAM/CRAM/VCF/dump format"""

    def __init__(
        self,
        gene: Gene,
        profile: Optional[Profile],
        path: str,
        reference: Optional[str] = None,
        debug: Optional[str] = None,
    ):
        """
        :param gene: Gene instance.
        :param profile: Profile instance. `None` if loading a dump file.
        :param path: Path to a SAM/BAM/CRAM/VCF/dump file.
        :param reference: Reference genome path for reading CRAM files.
            Default: None.
        :param debug: When set, create a `{debug}.dump` file for debug purposes.
            Default: None.

        :raise: :py:class:`aldy.common.AldyException` if the sample is invalid
            (e.g., the coverage of the copy-number neutral region is too low).
        """

        self.name = os.path.basename(path).split(".")[0]
        """Sample name."""

        self.gene = gene
        """Gene instance."""

        self.profile = profile
        """Profile instance."""

        self.path = path
        """File path."""

        self._dump_cn = defaultdict(int)
        """dict[int, int]: Coverage of the copy-number neutral region"""

        self._dump_reads: List[Tuple[Tuple, List]] = []
        """list[tuple[tuple, list]]: Read information."""

        self._indel_sites = {
            (pos, op): [0, 0] for pos, op in gene.mutations if op[:3] in ["ins", "del"]
        }
        self._indel_sites_eqs = {}
        """indel equivalents (used for long reads because indelpost is too slow)"""

        self._multi_sites = {
            m.pos: m.op
            for _, a in gene.alleles.items()
            for m in a.func_muts
            if ">" in m.op and len(m.op) > 3
        }
        """Multi-substitutions (e.g., `A.C>T.G`)."""

        self.phaseable = {
            pos: i for i, pos in enumerate(sorted({pos for pos, _ in gene.mutations}))
        }
        """Locations that should be phased."""

        self.phases: Dict[str, Dict[int, str]] = {}
        """Phasing information."""

        self._fusion_counter: Dict = {}
        """Fusion read coverage (for long reads)."""

        self.is_long_read = False
        """Set if long-read data is used."""

        with Timing("[sam] Read SAM"):
            self.kind, _ = detect_genome(path)
            self.genome = gene.genome

            if self.kind == "vcf":
                try:
                    norm, muts = self._load_vcf(path)
                except ValueError:
                    raise AldyException(f"VCF {path} is not indexed")
            elif self.kind == "dump":
                norm, muts = self._load_dump(path)
            else:
                if self.profile and self.profile.sam_long_reads:
                    self.is_long_read = True
                    norm, muts = self._load_long_sam(path, reference, debug)
                else:
                    norm, muts = self._load_sam(path, reference, debug)
                if self.profile and self.profile.cn_region:
                    self._dump_cn = self._load_cn_region(
                        path, reference, self.profile.cn_region
                    )
            self._make_coverage(norm, muts)
            if self.kind == "sam" and debug:
                self._dump_alignments(f"{debug}.{gene.name}", norm, muts)

        assert self.profile, "profile not set"
        if self.profile.cn_region:
            self.coverage._normalize_coverage()
        log.debug("[sam] avg_coverage= {:.1f}x", self.coverage.average_coverage())
        if self.profile.cn_region and self.coverage.diploid_avg_coverage() < 2:
            raise AldyException(
                "The average coverage of the sample is too low ({:.1f}).".format(
                    self.coverage.diploid_avg_coverage()
                )
            )

    def _load_sam(self, sam_path: str, reference=None, debug=None):
        """
        Load the read, mutation and coverage data from a SAM/BAM/CRAM file.
        :raise: :py:class:`aldy.common.AldyException` if the file lacks an index.
        """

        log.debug("[sam] path= {}", os.path.abspath(sam_path))
        if reference:
            log.debug("[sam] reference= {}", os.path.abspath(reference))
        assert self.profile, "profile not set"

        norm: dict = defaultdict(list)
        muts: dict = defaultdict(list)

        with pysam.AlignmentFile(  # type: ignore
            sam_path, reference_filename=reference
        ) as sam:
            # Check do we have proper index to speed up the queries
            try:
                has_index = sam.check_index()
            except AttributeError:
                # SAM files do not have an index. BAMs might also lack it
                has_index = False
            except ValueError:
                raise AldyException(f"Cannot check index of {sam_path}")

            self._prefix = chr_prefix(
                self.gene.chr, [x["SN"] for x in sam.header["SQ"]]
            )

            with tempfile.TemporaryDirectory() as tmp:
                self._realign_indels(tmp, sam, reference)
            if has_index:
                iter = sam.fetch(
                    region=self.gene.get_wide_region().samtools(prefix=self._prefix)
                )
            else:
                log.warn("SAM/BAM index not found. Reading will be slow.")
                iter = sam.fetch()
            # Fetch the reads
            for read in iter:
                if not read.cigartuples:  # only valid alignments
                    continue
                if read.is_supplementary:  # avoid supplementary alignments
                    continue
                if "H" in read.cigarstring:  # avoid hard-clipped reads
                    continue
                if not read.query_sequence:
                    continue
                # ensure that it is a proper gene read
                if not _in_region(self.gene.get_wide_region(), read, self._prefix):
                    continue

                # Handle 10X Genomics tags
                if read.has_tag("BX"):
                    fragment = read.get_tag("BX")
                    if read.has_tag("XC"):
                        fragment = f"{fragment}:{read.get_tag('XC')}"
                    elif read.has_tag("MI"):
                        fragment = f"{fragment}:{read.get_tag('MI')}"
                    self.is_long_read = True
                else:
                    fragment = read.query_name
                r = self._parse_read(
                    fragment,
                    read.reference_start,
                    read.cigartuples,
                    read.query_sequence,
                    norm,
                    muts,
                    read.mapping_quality,
                    read.query_qualities,
                )
                if r and debug:
                    self._dump_reads.append(r)
        return norm, muts

    def _load_vcf(self, vcf_path: str):
        """Load the read, mutation and coverage data from a VCF file."""

        log.debug("[vcf] path= {}", os.path.abspath(vcf_path))

        norm = {
            p: [(40, 40)] * 20
            for p in range(
                self.gene.get_wide_region().start - 500,
                self.gene.get_wide_region().end + 1,
            )
        }
        muts: dict = defaultdict(list)

        def get_mut(pos, ref, alt):
            off = 0
            while off < len(ref) and off < len(alt) and ref[off] == alt[off]:
                off += 1
            if len(ref) - off == 1 and len(alt) - off == 1:
                if alt[off] == self.gene[off + pos]:
                    return off + pos, "_"
                return off + pos, f"{self.gene[off + pos]}>{alt[off]}"
            elif len(ref) > len(alt) and len(alt) - off == 0:
                return off + pos, f"del{self.gene[off + pos : pos + len(ref)]}"
            elif len(ref) < len(alt) and len(ref) - off == 0:
                return off + pos, f"ins{alt[off:]}"
            else:
                log.trace(f"[sam] ignoring {pos}: {ref}->{alt}")
                return pos, None

        with pysam.VariantFile(vcf_path) as vcf:  # type: ignore
            self._prefix = chr_prefix(self.gene.chr, list(vcf.header.contigs))

            samples = list(vcf.header.samples)
            self.name = sample = samples[0]
            if len(samples) > 1:
                log.warn("WARNING: Multiple VCF samples found; using the first one.")
            log.info("Using VCF sample {}", sample)
            for read in vcf.fetch(
                region=self.gene.get_wide_region().samtools(prefix=self._prefix)
            ):
                g = sorted(y for y in read.samples[sample]["GT"] if y is not None)
                if len(g) != 2 or self.gene[read.pos - 1] == "N":
                    continue  # ignore polyploid and incomplete cases
                dump_arr = {}
                if len(read.ref) == 1 and read.ref != self.gene[read.pos - 1]:
                    hgvs = [(read.pos - 1, f"{self.gene[read.pos - 1]}>{read.ref}")]
                else:
                    hgvs = [(read.pos - 1, "_")]
                hgvs += [get_mut(read.pos - 1, read.ref, a) for a in read.alleles[1:]]
                for gt in g:
                    pos, op = hgvs[gt]
                    if op == "_":
                        continue
                    muts[pos, op] += [(40, 40)] * 10
                    norm[pos] = norm[pos][:-10]
                    dump_arr[pos] = op

                # Handle multi-SNPs
                for pos, op in self._multi_sites.items():
                    if pos not in dump_arr:
                        continue
                    l, r = op.split(">")
                    if all(
                        dump_arr.get(pos + p, "-") == f"{l[p]}>{r[p]}"
                        for p in range(len(l))
                        if l[p] != "."
                    ):
                        for p in range(len(l)):
                            if l[p] != ".":
                                np = pos + p, f"{l[p]}>{r[p]}"
                                muts[np] = muts[np][:-10]
                                if p:
                                    norm[pos + p] += [(40, 40)] * 10
                        muts[pos, op] += [(40, 40)] * 10
        return norm, muts

    def _load_dump(self, dump_path: str):
        """Load Aldy dump data."""

        log.debug("[dump] path= {}", os.path.abspath(dump_path))
        if dump_path.endswith(".tar.gz"):
            tar = tarfile.open(dump_path, "r:gz")

            f = [i for i in tar.getnames() if i.endswith(f".{self.gene.name}.dump")]
            if not f:
                raise AldyException("Invalid dump file")
            log.debug("Found {} in the archive", f[0])
            data = tar.extractfile(f[0])
            assert data, "Malformed dump file"
            fd = gzip.open(data)
        else:
            fd = gzip.open(dump_path, "rb")

        log.warn("Loading debug dump from {}", dump_path)
        (
            self.name,
            self.profile,
            self._dump_cn,
            norm,
            muts,
            phases,
            self._fusion_counter,
            self._indel_sites,
        ) = pickle.load(
            fd
        )  # type: ignore
        self.phases = {f"r{i}": v for i, v in enumerate(phases)}
        norm = {p: [q for q, n in c.items() for _ in range(n)] for p, c in norm.items()}
        muts = {p: [q for q, n in c.items() for _ in range(n)] for p, c in muts.items()}
        return norm, muts

    def _dump_alignments(self, debug: str, norm, muts):
        """Pickle alignment data for debug purposes."""
        with open(f"{debug}.genome", "w") as fd:
            print(self.gene.genome, file=fd)
        with gzip.open(f"{debug}.dump", "wb") as fd:  # type: ignore
            pickle.dump(
                (
                    self.name,
                    self.profile,
                    self._dump_cn,
                    {p: Counter(q) for p, q in norm.items()},
                    {p: Counter(q) for p, q in muts.items()},
                    [v for v in self.phases.values() if len(v) > 1],
                    self._fusion_counter,
                    self._indel_sites,  # TODO: remove
                ),
                fd,  # type: ignore
            )

    def _realign_indels(self, tmp, sam, reference, long_reads=False):
        """Realign reads around database indels via indelpost module."""

        assert self.profile, "profile not loaded"

        rname = f"{self._prefix}{self.gene.chr}"
        if not reference:
            reference = f"{tmp}/ref.fa"
            with open(reference, "w") as f:
                print(f">{rname}", file=f)
                print("N" * self.gene._lookup_range[0], end="", file=f)
                print(self.gene._lookup_seq, end="", file=f)
                sz = sam.get_reference_length(rname)
                print("N" * (sz - self.gene._lookup_range[1]), file=f)
            with open(f"{reference}.fai", "w") as f:
                print(rname, sz, len(rname) + 2, sz, sz + 1, sep="\t", file=f)
        ref = pysam.FastaFile(reference)  # type: ignore

        prev_indel = None
        for pos, op in sorted(self._indel_sites, key=lambda x: (x[0], -len(x[1]))):
            p = pos
            if op.startswith("del"):
                if "ins" in op:
                    pd, pi = op[3:].split("ins")
                    o1, o2 = pd, pi
                else:
                    p -= 1
                    o = self.gene[p]
                    o1, o2 = o + op[3:], o
            else:
                o = self.gene[p]
                o1, o2 = o, o + op[3:]

            v = Variant(rname, p + 1, o1, o2, ref)  # type: ignore

            if long_reads:
                # Speed-up: just generate equivalent indels, no need for the realignment
                for ev in v.generate_equivalents():
                    np, no = ev.pos - 1, ""
                    if len(ev.ref) < len(ev.alt) and ev.alt.startswith(ev.ref):
                        np += len(ev.ref)
                        no = "ins" + ev.alt[len(ev.ref) :]
                    elif len(ev.ref) > len(ev.alt) and ev.ref.startswith(ev.alt):
                        np += len(ev.alt)
                        no = "del" + ev.ref[len(ev.alt) :]
                    if no:
                        self._indel_sites_eqs[np, no] = (pos, op)
                continue

            exact_match_for_shiftable = True
            valn = VariantAlignment(  # type: ignore
                v,
                sam,
                mapping_quality_threshold=self.profile.min_mapq,
                base_quality_threshold=self.profile.min_quality,
                # needed to account for indel and database errors
                exact_match_for_shiftable=exact_match_for_shiftable,
            )

            phased = valn.phase()
            if len(phased.ref) - len(phased.alt) != len(v.ref) - len(v.alt):
                continue  # HACK: this indicates a subsumed indel

            self._indel_sites[pos, op] = list(valn.count_alleles())
            on_target = {r.query_name for r in valn.fetch_reads("target")}
            if (
                prev_indel
                and pos == prev_indel[0]
                and op.startswith("ins")
                and prev_indel[1].startswith(op)  # handle indel repeats
            ):
                # do not consider previously used reads
                off, on = self._indel_sites[pos, op]
                x = len(on_target & prev_indel[2])
                self._indel_sites[pos, op] = [off + x, on - x]
            if self._indel_sites[pos, op][1]:
                log.debug(
                    "[indel] {}:{} -> {}", pos + 1, op, self._indel_sites[pos, op]
                )
                prev_indel = (pos, op, on_target)
        # assert False

    def _load_cn_region(self, path, reference, cn_region):
        """
        Load the copy-number neutral coverage data from a SAM/BAM/CRAM file.

        :param cn_region: Copy-number neutral region to be used for coverage rescaling.
        """
        self._dump_cn = defaultdict(int)
        with pysam.AlignmentFile(  # type: ignore
            path, reference_filename=reference
        ) as sam:
            # Check do we have proper index to speed up the queries
            try:
                has_index = sam.check_index()
            except AttributeError:
                # SAM files do not have an index. BAMs might also lack it
                has_index = False
            except ValueError:
                raise AldyException(f"Cannot check index of {self.path}")
            self._prefix = chr_prefix(
                self.gene.chr, [x["SN"] for x in sam.header["SQ"]]
            )
            # Set it to _fetched_ if a CN-neutral region is user-provided.
            # Then read the CN-neutral region.
            if has_index:
                iter = sam.fetch(region=cn_region.samtools(prefix=self._prefix))
            else:
                log.warn("SAM/BAM index not found. Reading will be slow.")
                iter = sam.fetch()
            for read in iter:
                if _in_region(cn_region, read, self._prefix):
                    start = read.reference_start
                    if read.cigartuples is None:
                        continue
                    if read.is_supplementary:
                        continue
                    for op, size in read.cigartuples:
                        if op in [0, 7, 8, 2]:
                            for i in range(size):
                                self._dump_cn[start + i] += 1
                            start += size
        return self._dump_cn

    def _make_coverage(self, norm, muts):
        """Populate coverage data."""

        coverage: Dict[int, Dict[str, List]] = dict()
        for pos, cov in norm.items():
            if len(cov) == 0:
                continue
            coverage.setdefault(pos, {})["_"] = cov
        bounds = min(self.gene.chr_to_ref), max(self.gene.chr_to_ref)
        for (pos, mut), cov in muts.items():
            if pos not in coverage:
                coverage[pos] = {}
            if not bounds[0] <= pos <= bounds[1] and mut[:3] != "ins":
                mut = "_"  # ignore mutations outside of the region of interest
            coverage.setdefault(pos, {}).setdefault(mut, []).extend(cov)
        for pos, op in self._multi_sites.items():
            if pos in coverage and op in coverage[pos]:
                log.debug(f"[sam] multi-SNP {pos}{op}: {len(coverage[pos][op])} reads")
        assert self.profile, "profile not set"
        self.coverage = Coverage(
            self.gene,
            self.profile,
            self,
            {p: {m: v for m, v in coverage[p].items() if len(v) > 0} for p in coverage},
            self._indel_sites,
            self._dump_cn,
        )
        """Sample coverage data."""

        return self.coverage

    def _parse_read(
        self,
        fragment,
        ref_start,
        cigar,
        seq,
        norm,
        muts,
        mq=None,
        qual=None,
    ):
        """
        Parse a read.

        :param gene: Gene instance.
        :param fragment: Read name.
        :param ref_start: Start position in the reference genome.
        :param cigar: List of CIGAR operations in tuple format (operation, size).
        :param seq: Read sequence.
        :param norm: Positions within the read that have not been mutated.
        :param muts: Positions within the read that have been mutated.
        :param mq: Mapping quality.
        :param seq: Read qualities (normalized).

        .. note:: `norm` and `muts` are modified.

        .. note:: Qualities will be binned.
        """

        def bin_quality(q):
            """
            Quality score binning.
            See https://www.illumina.com/content/dam/illumina-marketing/docum ents/products/technotes/technote_understanding_quality_scores.pdf  # noqa
            """
            if q < 2:
                return int(q)
            if q < 10:
                return 6
            if q < 20:
                return 15
            if q < 29:
                return 25
            if q < 39:
                return 35
            return 40

        phase = self.phases.setdefault(fragment, {})
        dump_arr = []
        start, s_start = ref_start, 0
        prev_q = 10
        for op, size in cigar:
            if op == 2:  # Deletion
                mut = (start, "del" + self.gene[start : start + size])
                for i in range(size):
                    muts[start + i, "-"].append((bin_quality(mq), bin_quality(prev_q)))
                dump_arr.append(mut)
                if start in self.phaseable:
                    phase[start] = mut[1]
                if self._indel_sites_eqs and mut in self._indel_sites_eqs:
                    self._indel_sites[self._indel_sites_eqs[mut]][1] += 1
                start += size
            elif op == 1:  # Insertion
                mut = (start, "ins" + seq[s_start : s_start + size])
                q = mean(qual[s_start : s_start + size]) if qual else prev_q
                muts[mut].append((bin_quality(mq), bin_quality(q)))
                prev_q = q
                dump_arr.append(mut)
                if start in self.phaseable:
                    phase[start] = mut[1]
                if self._indel_sites_eqs and mut in self._indel_sites_eqs:
                    self._indel_sites[self._indel_sites_eqs[mut]][1] += 1
                s_start += size
            elif op == 4:  # Soft-clip
                s_start += size
            elif op in [0, 7, 8]:  # M, X and =
                for i in range(size):
                    q = qual[s_start + i] if qual else prev_q
                    if (
                        start + i in self.gene
                        and self.gene[start + i] != seq[s_start + i]
                    ):
                        mut = (start + i, f"{self.gene[start + i]}>{seq[s_start + i]}")
                        dump_arr.append(mut)
                        muts[mut].append((bin_quality(mq), bin_quality(q)))
                        if start + i in self.phaseable:
                            phase[start + i] = mut[1]
                    else:  # We ignore all mutations outside the RefSeq region
                        norm[start + i].append((bin_quality(mq), bin_quality(q)))
                        if start + i in self.phaseable:
                            phase[start + i] = "_"
                    prev_q = q
                start += size
                s_start += size

        dump_arr_pos = {p for p, _ in dump_arr}
        for pos, op in self._multi_sites.items():
            if pos not in dump_arr_pos:
                continue
            l, r = op.split(">")
            if all(
                (pos + p, f"{l[p]}>{r[p]}") in dump_arr
                for p in range(len(l))
                if l[p] != "."
            ):
                items = []
                for p in range(len(l)):
                    if l[p] != ".":
                        items.append(muts[pos + p, f"{l[p]}>{r[p]}"].pop())
                        if p:  # no idea why...
                            norm[pos + p].append(items[-1])
                # TODO: use sth else instead of mean?
                muts[pos, op].append(
                    (mean(mq for mq, _ in items), mean(q for _, q in items))
                )

        if self._indel_sites_eqs:  # long-read hack
            for pos, op in self._indel_sites:
                if ref_start <= pos < start:
                    self._indel_sites[pos, op][0] += 1
        read_pos = (ref_start, start, len(seq))
        return read_pos, dump_arr

    def _load_long_sam(self, sam_path: str, reference=None, debug=None):
        """
        Load the read, mutation and coverage data from a long-read SAM/BAM/CRAM file.
        This function remaps long reads for selected genes.
        Tested on PacBio HiFi data only.

        .. note:: Currently only supports CYP2D6.
        """

        log.debug("[sam] pacbio_path= {}", os.path.abspath(sam_path))
        assert self.genome, "Genome not provided"
        assert self.genome == "hg38", "Only hg38 supported at this moment"
        assert self.profile, "profile not set"

        norm: dict = defaultdict(list)
        muts: dict = defaultdict(list)

        self._index = None
        ref = script_path(f"aldy.resources.genes/{self.gene.name.lower()}.fa.gz")
        if os.path.exists(ref):
            import mappy

            self._index = mappy.Aligner(ref, preset=self.profile.sam_mappy_preset)

            seq_name = self._index.seq_names[0]
            self._seq_range = list(map(int, seq_name.split(":")[1].split("-")))
            self._index_gene = []
            for _, g in enumerate(self.gene.regions):
                st = min(r.start for r in g.values())
                ed = max(r.end for r in g.values())
                idx = mappy.Aligner(
                    seq=self._index.seq(seq_name)[
                        st - self._seq_range[0] : ed - self._seq_range[0]
                    ],
                    preset=self.profile.sam_mappy_preset,
                )
                self._index_gene.append((idx, st))

            self._fusion_signatures = {}
            self._fusion_counter = {}
            for n, cn in self.gene.cn_configs.items():
                if cn.kind == CNConfigType.LEFT_FUSION:
                    s, pos = 0, next(i for i, j in cn.cn[0].items() if j == 1)
                elif cn.kind == CNConfigType.RIGHT_FUSION:
                    s, pos = 1, next(i for i, j in cn.cn[0].items() if j == 0)
                else:
                    continue
                sig = (
                    (1 - s, list(cn.cn[0])[list(cn.cn[0]).index(pos) - 1]),
                    (s, pos),
                )
                self._fusion_counter[n] = [0, 0]
                if (
                    sig[0][1] in self.gene.unique_regions
                    or sig[1][1] in self.gene.unique_regions
                ):
                    self._fusion_signatures[sig] = n

            log.debug("[sam] PacBio remapping enabled")

        with pysam.AlignmentFile(  # type: ignore
            sam_path, reference_filename=reference
        ) as sam, Timing("[sam] Remap"):
            # Assumes SAM index exists
            self._prefix = chr_prefix(
                self.gene.chr, [x["SN"] for x in sam.header["SQ"]]
            )

            with tempfile.TemporaryDirectory() as tmp:
                self._realign_indels(tmp, sam, reference, True)
                for po, (off, on) in self._indel_sites.items():
                    self._indel_sites[po] = [off - on, on]

            wide = self.gene.get_wide_region()
            end = wide.end
            if self.gene.name == "CYP2D6":
                end = {"hg19": 42_551_000, "hg38": 42_155_000}[self.genome]
            rng = GRange(wide.chr, wide.start, end)
            counter = 0
            iter = sam.fetch(region=rng.samtools(prefix=self._prefix))
            for read in iter:
                if not read.cigartuples or "H" in read.cigarstring:
                    continue
                if read.reference_start >= rng.end or read.reference_end < rng.start:
                    continue

                seq = read.query_sequence
                qual = read.query_qualities if read.query_qualities else [40] * len(seq)
                if self._index:
                    # Split reads
                    pieces = []
                    s_start = 0
                    gene_bnd = self.gene.regions[0]["up"].end - self._seq_range[0]
                    pseudo_bnd = self.gene.regions[1]["up"].end - self._seq_range[0]

                    while s_start < len(seq):
                        s = seq[s_start:]
                        q = qual[s_start:]
                        h = self._map(self._index, s)
                        if (
                            h
                            and h.r_st <= gene_bnd < h.r_en
                            and gene_bnd - h.r_st > 100
                        ):
                            s = seq[s_start : s_start + (gene_bnd - h.r_st)]
                            q = qual[s_start : s_start + (gene_bnd - h.r_st)]
                            h = self._map(self._index, s)
                        elif (
                            h
                            and h.r_st <= pseudo_bnd < h.r_en
                            and pseudo_bnd - h.r_st > 100
                        ):
                            s = seq[s_start : s_start + (pseudo_bnd - h.r_st)]
                            q = qual[s_start : s_start + (pseudo_bnd - h.r_st)]
                            h = self._map(self._index, s)
                        if not h:
                            s_start += 100
                        else:
                            pcs, ed = self._split_read(s, q, h)
                            pieces += pcs
                            s_start += ed
                    for ref_start, seq, qual, cigar in pieces:
                        r = self._parse_read(
                            counter,
                            ref_start,
                            cigar,
                            seq,
                            norm,
                            muts,
                            read.mapping_quality,
                            qual,
                        )
                        counter += 1
                        if r and debug:
                            self._dump_reads.append(r)
                else:
                    r = self._parse_read(
                        counter,
                        read.reference_start,
                        read.cigartuples,
                        seq,
                        norm,
                        muts,
                        read.mapping_quality,
                        read.query_qualities,
                    )
                    counter += 1
                    if r and debug:
                        self._dump_reads.append(r)
        return norm, muts

    def _map(self, idx, seq):
        """
        Map a sequence to a mappy reference index.
        :returns: The first primary hit or `None` if no primary hits are found.
        """

        for hit in idx.map(seq):
            if hit.is_primary:
                return hit
        return None

    def _split_read(self, seq, qual, hit):
        """Split a long read."""

        def _rev_cigar(c):
            """Reverse CIGAR tuples for pysam/mappy compatibility."""
            return [(op, size) for (size, op) in c]

        regs = self._get_gene_regions(
            hit.r_st + self._seq_range[0] - 1, _rev_cigar(hit.cigar)
        )
        if regs:
            splits = []
            for fs, fn in self._fusion_signatures.items():
                brk = [r for r in regs if r[0][1] == fs[1][1]]
                if not brk:
                    continue
                self._fusion_counter[fn][1] += 1
                brk = brk[0]
                lg, brk, rg = fs[1][0], brk[1] + brk[2], fs[0][0]
                lh = self._map(self._index_gene[lg][0], seq[:brk])
                rh = self._map(self._index_gene[rg][0], seq[brk:])
                if lh and rh:
                    unmap = len(seq) - (lh.q_en - lh.q_st) - (rh.q_en - rh.q_st)
                    unmap += lh.NM + rh.NM
                    splits.append([fn, unmap, brk, lh, rh, lg, rg])
            splits.sort(key=lambda x: x[1])
            if splits and splits[0][1] <= hit.NM:
                fn, _, brk, lh, rh, lg, rg = splits[0]
                self._fusion_counter[fn][0] += 1
                return (
                    [
                        (
                            lh.r_st + self._index_gene[lg][1] - 1,
                            seq[lh.q_st : lh.q_en],
                            qual[lh.q_st : lh.q_en],
                            _rev_cigar(lh.cigar),
                        ),
                        (
                            rh.r_st + self._index_gene[rg][1] - 1,
                            seq[brk + rh.q_st : brk + rh.q_en],
                            qual[brk + rh.q_st : brk + rh.q_en],
                            _rev_cigar(rh.cigar),
                        ),
                    ],
                    max(lh.q_en, brk + rh.q_en),
                )
        return [
            (
                hit.r_st + self._seq_range[0] - 1,
                seq[hit.q_st : hit.q_en],
                qual[hit.q_st : hit.q_en],
                _rev_cigar(hit.cigar),
            )
        ], hit.q_en

    def _get_gene_regions(self, r_start, cigar):
        """Get gene regions that cover the given CIGAR range."""

        regs = []
        start, s_start = r_start, 0
        for op, size in cigar:
            if op == 1 or op == 4:
                s_start += size
            elif op == 2 or op == 3:
                start += size
            elif op == 0 or op == 7 or op == 8:
                for i in range(size):
                    rg = self.gene.region_at(start + i)
                    if rg:
                        if not regs or regs[-1][0] != rg:
                            regs.append([rg, s_start + i, 0])
                        regs[-1][2] += 1
                start += size
                s_start += size
        return regs


def detect_genome(sam_path: str) -> Tuple[str, Optional[str]]:
    """Detect file type and the reference genome."""

    try:
        pysam.set_verbosity(0)  # type: ignore
        with pysam.AlignmentFile(sam_path) as sam:  # type: ignore
            try:
                sam.check_index()
            except AttributeError:
                pass  # SAM files do not have an index. BAMs might also lack it
            except ValueError:
                raise AldyException(
                    f"File {sam_path} has no index (it must be indexed)"
                )

            #: str: Check if we need to append 'chr' or not
            _prefix = chr_prefix("1", [x["SN"] for x in sam.header["SQ"]])
            # Detect the genome
            hg19_lens = {"1": 249250621, "10": 135534747, "22": 51304566}
            hg38_lens = {"1": 248956422, "10": 133797422, "22": 50818468}
            chrs = {x["SN"]: int(x["LN"]) for x in sam.header["SQ"]}
            is_hg19, is_hg38 = True, True
            for c in "1 10 22".split():
                if _prefix + c in chrs:
                    is_hg19 &= chrs[_prefix + c] == hg19_lens[c]
                    is_hg38 &= chrs[_prefix + c] == hg38_lens[c]
            if is_hg19:
                return "sam", "hg19"
            elif is_hg38:
                return "sam", "hg38"
            else:
                return "sam", None
    except (ValueError, OSError):
        try:
            with pysam.VariantFile(sam_path):  # type: ignore
                return "vcf", None
        except (ValueError, OSError):
            if sam_path.endswith(".tar.gz"):
                tar = tarfile.open(sam_path, "r:gz")
                f = [i for i in tar.getnames() if i.endswith(".genome")]
                if not f:
                    raise AldyException("Invalid dump file")
                data = tar.extractfile(f[0])
                if data:
                    genome = data.read().decode("utf-8").strip()
                    return "dump", genome
    return "", None


def _in_region(region: GRange, read, prefix: str) -> bool:
    """Check if a read intersects a given gene region."""

    if read.reference_id == -1 or read.reference_name != prefix + region.chr:
        return False
    if read.reference_end is None:
        return False

    a = (read.reference_start, read.reference_end)
    b = (region.start, region.end)
    return a[0] <= b[0] <= a[1] or b[0] <= a[0] <= b[1]



# 786
# Aldy source: coverage.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, Tuple, Callable, List, Any, Optional
import copy

from .profile import Profile
from .common import log, AldyException
from .gene import Mutation, Gene
from .solutions import CNSolution


class Coverage:
    """Data structure that maintains the coverage information for a given sample."""

    def __init__(
        self,
        gene: Gene,
        profile: Profile,
        sam,
        coverage: Dict[int, Dict[str, List]],
        indel_coverage: Optional[Dict],
        cnv_coverage: Dict[int, int],
    ) -> None:
        """
        :param gene: Gene instance.
        :param profile: Profile instance.
        :param sam: Sample instance.
        :param coverage: Coverage for each sample location. Each location is represented
            as a dictionary that maps a mutation (or a reference position indicated by
            `_`) to the list of read quality scores that cover it. For example,
            `coverage[10]['A>G'] = [(10, 20), (10, 10)]` indicates that 2 reads have G
            (instead of A) at the location 10.
        :param indel_coverage: Number of reads that do not support and that do support
            the indel for each indel in the gene database.
        :param cnv_coverage: Coverage of the copy-number neutral region within the
            sample. Each location is represented by the total corresponsing read
            coverage. Used for coverage rescaling.
        """
        self.gene = gene
        self.profile = profile
        self.sam = sam
        self._coverage = {}
        for pos, ops in coverage.items():
            self._coverage[pos] = {}
            for op, quals in ops.items():
                if not (indel_coverage and op.startswith("ins")):
                    self._coverage[pos][op] = quals
        if indel_coverage:
            self._indels = {k: (n, y) for k, (n, y) in indel_coverage.items() if y}
        else:
            self._indels = None
        self._cnv_coverage = cnv_coverage
        self._region_coverage: Dict[Tuple[int, str], float] = {}

    def __getitem__(self, mut: Mutation) -> float:
        """:returns: Mutation coverage."""
        return self.coverage(mut)

    def coverage(self, mut: Mutation) -> float:
        """:returns: Mutation coverage."""
        if self._indels and (mut.pos, mut.op) in self._indels:
            return self._indels[mut.pos, mut.op][1]
        if mut.pos in self._coverage and mut.op in self._coverage[mut.pos]:
            return len(self._coverage[mut.pos][mut.op])
        else:
            return 0

    def total(self, m) -> float:
        """:returns: Location coverage."""
        if hasattr(m, "pos") and hasattr(m, "op"):
            if self._indels and (m.pos, m.op) in self._indels:
                return sum(self._indels[m.pos, m.op])
            pos = m.pos
        else:
            assert isinstance(m, int), f"got {m}: {type(m)} {Mutation}"
            pos = m
        if pos not in self._coverage:
            return 0
        return float(
            sum(len(v) for p, v in self._coverage[pos].items() if p[:3] != "ins")
        )

    def percentage(self, m: Mutation) -> float:
        """:returns: Mutation coverage expressed as percentage (0-100%)."""
        total = self.total(m)
        if total == 0:
            return 0
        return 100.0 * self.coverage(m) / total

    def single_copy(self, m, cn_solution: CNSolution) -> float:
        """
        :param pos: Genomic locus.
        :param cn_solution: Copy-number solution.
        :returns: Coverage of a *single* gene copy at the given location.
        """
        if hasattr(m, "pos"):
            pos = m.pos
        else:
            assert isinstance(m, int)
            pos = m
        if cn_solution.position_cn(pos) == 0:
            return 0
        return max(1, self.total(m)) / cn_solution.position_cn(pos)

    def region_coverage(self, gene: int, region: str) -> float:
        """:returns: Average coverage of a gene region."""
        return self._region_coverage[gene, region]

    def average_coverage(self) -> float:
        """:returns: Average coverage of the gene."""
        return sum(self.total(pos) for pos in self._coverage) / float(
            len(self._coverage) + 0.1
        )

    def dump(self, out=None):
        """Pretty-print the coverage data."""
        for pos, pos_mut in sorted(self._coverage.items()):
            if len(pos_mut) == 1 and "_" in pos_mut:
                continue
            for _, (op, _) in enumerate(sorted(pos_mut.items(), reverse=True)):
                p = self.percentage(Mutation(pos, op))
                if pos in self.gene:
                    x = self.gene.get_functional((pos, op))
                    if op == "_":
                        x = ""
                    if x and (pos, op) not in self.gene.mutations:
                        x += "**"
                    r = self.gene.region_at(pos)
                    t = f"{x if x else ''}\t{r[1] if r else ''}"
                    t += "\t" + self.gene.get_rsid((pos, op))
                else:
                    t = "\t\t"
                if out and self.sam:
                    out(
                        f"[dump] {self.sam.name}\t{self.gene.name}\t"
                        f"{self.gene.chr_to_ref.get(pos, -1) + 1}\t{op}\t{p:.1f}\t{t}"
                    )

    def filtered(self, filter_fn: Callable[[Any, Mutation], List]):
        """
        :param filter_fn: Function that performs mutation filtering with the following
            arguments:

                1. mut (:py:class:`aldy.gene.Mutation`): mutation to be filtered
                2. cov (float): coverage of the mutation
                3. total (float): total coverage of the mutation locus
                4. thres (float): filtering threshold

            `filter_fn` returns `False` if a mutation should be filtered out.

        :returns: Filtered coverage.
        """

        new_cov = copy.copy(self)
        new_cov._coverage = {}
        for pos, pos_mut in self._coverage.items():
            new_cov._coverage[pos] = {}
            for o in pos_mut:
                f = filter_fn(self, Mutation(pos, o))
                if isinstance(f, list) and f:
                    new_cov._coverage[pos][o] = f
                elif f:
                    assert isinstance(f, bool)
                    new_cov._coverage[pos][o] = pos_mut[o]
        new_cov._indels = None
        if self._indels:
            new_cov._indels = {}
            for (pos, o), v in self._indels.items():
                f = filter_fn(self, Mutation(pos, o))
                if not (isinstance(f, bool) and not f):
                    new_cov._indels[pos, o] = v
        return new_cov

    def diploid_avg_coverage(self) -> float:
        """:returns: Average coverage of the copy-number neutral region."""
        assert self.profile.cn_region, "CN region not set"
        return float(sum(self._cnv_coverage.values())) / abs(
            self.profile.cn_region.end - self.profile.cn_region.start
        )

    def _normalize_coverage(self) -> None:
        """Normalize the sample coverage with the profile coverage."""

        assert self.profile.cn_region and self.profile.data, "CN region not set"
        sam_ref = sum(
            self._cnv_coverage[i]
            for i in range(self.profile.cn_region.start, self.profile.cn_region.end)
        )
        if sam_ref == 0:
            raise AldyException(
                f"CN-neutral region {self.profile.cn_region} has no reads. "
                + "Double check your input file for CYP2D8 (are you using hg19?), "
                + "or pass an alternative CN-neutral region via -n parameter."
            )
        ratio = self.profile.neutral_value / sam_ref
        if ratio == 0:
            raise AldyException("Invalid CN-neutral region in the provided profile.")
        log.debug("[coverage] scale_ratio: {:.1f}", 1 / ratio)

        self._region_coverage = {}
        for gene, gr in enumerate(self.gene.regions):
            for region, rng in gr.items():
                s = sum(self.total(i) for i in range(rng.start, rng.end))
                p = self.profile.data[self.gene.name][region][gene]
                p /= 2  # profile has 2 copies, so divide it with 2 for normalization
                self._region_coverage[gene, region] = (ratio * s / p) if p != 0 else 0.0

    def basic_filter(self, mut: Mutation, cn=None, thres=None):
        """Basic threshold-based filter."""
        thres = (thres or self.profile.threshold) / (cn or 1)
        min_cov = max(self.profile.min_coverage, self.total(mut) * thres)
        # if self._indels and (mut.pos, mut.op) in self._indels:
        #     print(mut, self.coverage(mut), min_cov, thres, self.profile.threshold)
        sz = self.coverage(mut)
        return sz >= min_cov

    def quality_filter(self, mut: Mutation) -> List:
        """Basic quality filter."""
        quals = self._coverage.get(mut.pos, {}).get(mut.op, [])
        return [
            (m, q)
            for m, q in quals
            if q >= self.profile.min_quality
            if m >= self.profile.min_mapq
        ]



# cython: embedsignature=True

#from pysam.libcbcf cimport VariantRecord, VariantFile
from pysam.libcalignmentfile cimport AlignmentFile
from .variant cimport Variant
from .contig cimport Contig
from .local_reference cimport UnsplicedLocalReference

cdef class VariantAlignment:
    cdef Variant target, __target, second_target
    cdef readonly AlignmentFile bam
    cdef int window, retarget_window, mapqthresh
    cdef int downsamplethresh, basequalthresh, match_score, _observed_pos
    cdef int mismatch_penalty, gap_open_penalty, gap_extension_penalty
    cdef float retarget_cutoff, __sample_factor,
    cdef bint exclude_duplicates, exact_match_for_shiftable, auto_adjust_extension_penalty, no_realignment
    cdef list __pileup
    cdef readonly is_spurious_overhang, is_complex_input
    cdef readonly Contig contig
    cdef readonly UnsplicedLocalReference unspliced_local_reference

    cdef __parse_pileup(self, Contig contig=*, bint retargeted=*, bint skip_read_end_check=*)



#!/usr/bin/env python3

from .utilities cimport split
from .contig cimport Contig
from .variant cimport Variant
from .utilities import get_end_pos
from .consensus import is_compatible


def find_by_softclip_split(target, contig, pileup):
    """Annotate if reads contain target indel

    Args:
        target (Variant)
        template (dict): indel template
        pileup (list): list of dictized reads (dict)
    Return:
        pileup (list): annotated pileup
    """

    pos, indel_type, indel_seq = (target.pos, target.variant_type, target.indel_seq)

    pileup = [
        find_candidate_softclips(read, pos, indel_type, indel_seq) for read in pileup
    ]
    pileup = [
        is_target_by_sftclp_split(read, pos, indel_type, indel_seq, contig)
        for read in pileup
    ]

    return pileup


def find_candidate_softclips(read, pos, indel_type, indel_seq):
    """Find and annotate softclipped indels to be realigned

    Args:
        read (dict): dictized read with "is_target" key
        pos (int): 1-based
        indel_type (str): "I" for insertion "D" for deletion
        indel_seq (str): inserted or deleted sequence
    Returns:
        read (dict): dictized read with softclip info annotated.
    """

    # already identified as target or seq identical to reference
    if read["is_target"] or read["is_reference_seq"]:
        read["softclip_pattern"] = None
        return read

    # no softclipped bases
    if not "S" in read["cigar_string"]:
        read["softclip_pattern"] = None
        return read

    # softclipped reads covering the locus
    if read["is_covering"]:
        read["softclip_pattern"] = classify_softclip_patterns(read, pos)
        return read

    # reads with large deletion may not cover the locus
    if read["read_end"] < pos:
        if indel_type == "D" and pos < read["read_end"] + len(indel_seq):
            read["softclip_pattern"] = "trailing_deletion"
        else:
            read["softclip_pattern"] = None
    else:
        if indel_type == "D" and read["read_start"] - len(indel_seq) < pos:
            read["softclip_pattern"] = "leading_deletion"
        else:
            read["softclip_pattern"] = None

    return read


def classify_softclip_patterns(read, pos):
    """Annotate if softclip starts before or after indel postion

    Args:
        read (dict): dictized read
        pos (int): 1-based
    Returns:
        softclip pattern (str)
    """
    # event_pos = read["read_start"]  # 1-based genomic pos
    event_pos = read["covering_subread"][0]  # 1-based genomic pos

    last_event = "O"
    for i, c in enumerate(read["cigar_list"]):
        event, event_len = c[-1], int(c[:-1])
        event_pos += event_len

        if pos <= event_pos:
            last_event = event
            is_leading = i == 0
            break

    if last_event == "M":
        return "off_clipping"
    elif last_event == "S" and is_leading:
        return "leading"
    elif last_event == "S" and not is_leading:
        return "trailing"
    else:
        return "other"


def is_target_by_sftclp_split(read, pos, indel_type, indel_seq, contig, slided=False):
    """Find softclipped target indels

    Args:
        read (dict): dictized read
        pos (int): 1-based
        indel_type (str): "I" for insertion and "D" for deletion
        indel_seq (str): inserted or deleted sequence
        template (dict): indel template
        slided (bool): True if "slide_insertion(...)" applied
                       Used internally for recurrence
    Returns:
        read (dict):
    """
    if read["is_target"] or not read["softclip_pattern"]:
        return read

    read = split_softclipped_read(read, pos, indel_type, indel_seq)

    read["is_target"] = is_compatible(read, contig, indel_type)

    # slide insertions
    if not read["is_target"] and not slided and indel_type == "I":
        return is_target_by_sftclp_split(
            slide_insertion(read, contig),
            pos,
            indel_type,
            indel_seq,
            contig,
            slided=True,
        )

    if slided:
        read["read_start"] = read["orig_start"]
        read["read_end"] = read["orig_end"]

        del read["orig_start"], read["orig_end"]

    return read


def split_softclipped_read(read, pos, indel_type, indel_seq):

    indel_len = len(indel_seq)
    cigar_string = read["cigar_string"]
    reverse = True if read["softclip_pattern"] == "leading" else False
    string_pos = read["read_end"] if reverse else read["read_start"]

    if indel_type == "D" and reverse:
        pos += indel_len

    lt_flank, rt_flank = split(
        read["read_seq"],
        cigar_string,
        pos,
        string_pos,
        is_for_ref=False,
        reverse=reverse,
    )
    mid_seq = ""
    lt_qual, rt_qual = split(
        read["read_qual"],
        cigar_string,
        pos,
        string_pos,
        is_for_ref=False,
        reverse=reverse,
    )

    if indel_type == "I":
        mid_seq, rt_flank = rt_flank[:indel_len], rt_flank[indel_len:]
        read["del_seq"] = ""
    else:
        read["del_seq"] = indel_seq
        #del_pos = get_end_pos(read["read_start"], lt_flank, cigar_string)
        #del_pos = pos
        #aln_pos = read["aln_end"] if reverse else read["aln_start"]

        #lt_ref, rt_ref = split(
        #    read["ref_seq"],
        #    cigar_string,
        #    del_pos,
        #    aln_pos,
        #    is_for_ref=True,
        #    reverse=reverse,
        #)

        #read["del_seq"] = "GAATTAAGAGAAGCA"

    read["lt_flank"] = lt_flank
    read["lt_qual"] = lt_qual
    read["indel_seq"] = mid_seq
    read["rt_flank"] = rt_flank
    read["rt_qual"] = rt_qual

    return read


def slide_insertion(read, contig):
    """Move read by the length of insertion
       (this is for BWA aligned reads)

    Args:
        read (dict): dictized read
        template (dict): indel template
    Returns:
        read (dict): dictized read with slided position
    """
    total_slide = sum([int(c[:-1]) for c in contig.gaps if "I" in c])
    read["orig_start"] = read["read_start"]
    read["orig_end"] = read["read_end"]

    if read["softclip_pattern"] == "leading":
        read["read_start"] += total_slide
        read["softclip_pattern"] = "other"
    else:
        read["read_end"] -= total_slide
        read["softclip_pattern"] = "leading"

    return read



# 786
# Aldy source: genotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Optional, Any, Set, Dict
import os
import sys
import pkg_resources
import datetime
import time

from . import sam
from . import cn
from . import major
from . import minor
from . import diplotype
from . import solutions
from .common import (
    colorize,
    log,
    script_path,
    json,
    AldyException,
    SOLUTION_PRECISION,
)
from .profile import Profile
from .gene import Gene, GRange
from .diplotype import OUTPUT_COLS
from .lpinterface import model as lp_model


def genotype(
    gene_db: str,
    sam_path: str,
    profile_name: Optional[str],
    output_file: Optional[Any] = sys.stdout,
    cn_region: Optional[GRange] = None,
    cn_solution: Optional[List[str]] = None,
    solver: str = "any",
    reference: Optional[str] = None,
    debug: Optional[str] = None,
    multiple_warn_level: int = 1,
    report: bool = False,
    genome=None,
    is_simple: bool = False,
    **params,
) -> Dict[str, List[solutions.MinorSolution]]:
    """Genotype a sample.

    :param gene_db: Gene name (if it is shipped with Aldy)
        or the location of the gene database in YAML format.
    :param sam_path: Location of SAM/BAM/CRAM file that is to be analyzed.
    :param profile_name: Coverage profile (e.g. WGS).
        `None` if `cn_solution` is provided.
    :param output_file: Location of the output file. Use `None` for no output.
        Default: `sys.stdout`.
    :param cn_region: Copy-number neutral region.
        Default: None (uses the provided CYP2D8 region).
    :param cn_solution: List of the copy number configurations.
        Copy-number detection will not run if this parameter is set.
        Default: `None`.
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param reference: Reference genome (for reading CRAM files).
        Default: `None`.
    :param debug: Prefix for debug and core dump files.
        Default: `None` (no debug information).
    :param multiple_warn_level: Warning level (1 for optimal solutions, 2 for major
        solutions and 3 for CN solutions).
        Default: 1 (warn only on multiple optimal solutions).
    :param report: If set, write the solution summary to the stderr.
        Default: `False`.
    :param genome: Reference genome (e.g., hg19 or hg38).
        Default: `None` (auto-detect).
    :param is_simple: Use simple output format.
        Default: `False`.
    :param params: Model parameters. See :py:mod:`aldy.profile` for details.
    """

    t1 = time.time()
    log.debug("[genotype] gene={}; start={}", gene_db, datetime.datetime.now())

    # Test the existence of LP solver
    _ = lp_model("init", solver)

    with open(sam_path):  # Check if file exists
        pass
    kind, g = sam.detect_genome(sam_path)
    if genome is None:
        genome = g
        if not genome:
            log.warn("WARNING: Cannot detect genome, defaulting to hg19.")
            genome = "hg19"
        log.debug(f"[genotype] reference= {genome}")
    if genome not in ["hg19", "hg38"]:
        raise AldyException(f"Unknown genome {genome}")

    avail_genes = []
    if gene_db == "all":
        avail_genes = pkg_resources.resource_listdir("aldy.resources", "genes")
        avail_genes = [
            i[:-4]
            for i in avail_genes
            if len(i) > 4 and i.endswith(".yml") and not i.startswith("pharma-")
        ]
        avail_genes = sorted(avail_genes)
    elif gene_db == "pharmacoscan":
        avail_genes = pkg_resources.resource_listdir(
            "aldy.resources.genes", "pharmacoscan"
        )
        avail_genes = [
            f"pharmacoscan/{i[:-4]}" for i in avail_genes if i.endswith(".yml")
        ]
        avail_genes = sorted(avail_genes)
    else:
        avail_genes = gene_db.lower().split(",")
    if len(avail_genes) != 1:
        res: Dict = {}
        for a in avail_genes:
            log.warn("=" * 50)
            log.warn("Gene {}", a.upper())
            try:
                res = {
                    **res,
                    **genotype(
                        a,
                        sam_path,
                        profile_name,
                        output_file,
                        cn_region,
                        cn_solution,
                        solver,
                        reference,
                        debug,
                        multiple_warn_level,
                        report,
                        genome,
                        is_simple,
                        **params,
                    ),
                }
            except AldyException as ex:
                log.error(f"Failed gene {a.upper()}")
                log.error(f"Message: {str(ex)}")
            log.warn("")
        return res
    else:
        db_file = script_path(
            "aldy.resources.genes/{}.yml".format(avail_genes[0].lower())
        )

    # Load the gene specification
    if os.path.exists(db_file):
        gene_db = db_file
    with open(gene_db):  # Check if file exists
        pass
    gene = Gene(gene_db, genome=genome)

    if profile_name in ["exome", "wxs", "wes"]:
        gene.do_copy_number = False
        profile_name = "illumina"
        params["min_coverage"] = 5.0
    elif profile_name == "wgs":
        profile_name = "illumina"
    elif profile_name == "pgrnseq-v1":
        profile_name = "pgx1"
    elif profile_name == "pgrnseq-v2":
        profile_name = "pgx2"
    elif profile_name == "pgrnseq-v3":
        profile_name = "pgx3"

    if kind == "vcf":
        log.warn("WARNING: Using VCF file. Copy-number calling is not available.")
        profile = Profile("user_provided", cn_solution=["1", "1"], **params)
        sample = sam.Sample(gene, profile, sam_path, debug=debug)
    else:
        if cn_solution:
            profile = Profile("user_provided", cn_solution=cn_solution, **params)
        elif kind != "dump":
            if not profile_name:
                raise AldyException("Profile not provided")
            profile = Profile.load(gene, profile_name, cn_region, **params)
        else:
            profile = None
        sample = sam.Sample(gene, profile, sam_path, reference, debug)
    profile = sample.profile  # if loaded for a dump
    assert profile, "Profile not set"
    if kind == "dump":
        profile.update(params)

    json[gene.name].update({"sample": sample.name})
    is_vcf = output_file and output_file.name.endswith(".vcf")
    is_simple = is_simple or (
        output_file is not None and output_file.name.endswith(".simple")
    )
    is_aldy = output_file and not (is_vcf or is_simple)
    if is_simple:
        print(
            sample.name,
            gene.name,
            sep="\t",
            end="\t",
            file=output_file,
        )
    if kind != "vcf":
        avg_cov = sample.coverage.average_coverage()
        if profile.cn_region and avg_cov < 2:
            if is_simple:
                print(file=output_file)
            raise AldyException(
                f"Average coverage of {avg_cov:.2f} for gene {gene.name} is too low; "
                + f"skipping gene {gene.name}. "
                + f"Please ensure that {gene.name} is present in the input SAM/BAM."
            )
        elif profile.cn_region and avg_cov < 20:
            log.warn(
                f"Average sample coverage is {avg_cov}. "
                + "We recommend at least 20x coverage for the best results."
            )

    # Get copy-number solutions
    cn_sols = cn.estimate_cn(
        gene,
        profile,
        sample.coverage,
        solver=solver,
        debug=debug,
    )

    # Add SLACK to each score to avoid division by zero or other numerical issues
    # when the optimum is close to zero.
    # HACK: Value of 1 (1 gene copy) seems to be reasonable, but is not
    # theoretically guaranteed to be better than any other value.
    SLACK = 1

    # Get major solutions and pick the best one
    if multiple_warn_level >= 3 and len(cn_sols) > 1:
        log.warn("WARNING: multiple gene structures found!")
    log.info(f"Potential {gene.name} gene structures for {sample.name}:")
    major_sols: list = []
    cn_sols = sorted(cn_sols, key=lambda m: (int(1000 * m.score), m._solution_nice()))
    if len(cn_sols) == 0:
        if is_simple:
            print(file=output_file)
        raise AldyException("No solutions found!")
    min_cn_score = min(cn_sols, key=lambda m: m.score).score
    for i, cn_sol in enumerate(cn_sols):
        conf = 100 * (min_cn_score + SLACK) / (cn_sol.score + SLACK)
        log.info(f"  {i + 1:2}: {cn_sol._solution_nice()} (confidence: {conf:.0f}%)")
    log.debug("*" * 80)

    # return []

    for i, cn_sol in enumerate(cn_sols):
        sols = major.estimate_major(
            gene,
            sample.coverage,
            cn_sol,
            solver=solver,
            identifier=i,
            debug=debug,
        )
        for s in sols:
            s.score += cn_sol.score - min_cn_score
        major_sols += sols
    if len(major_sols) == 0:
        if is_simple:
            print(file=output_file)
        raise AldyException("No major solutions found!")

    major_sols = [
        solutions.MajorSolution(
            m.score,  # * ((m.cn_solution.score + SLACK) / (min_cn_score + SLACK)),
            m.solution,
            m.cn_solution,
            m.added,
        )
        for m in major_sols
    ]
    min_major_score = min(major_sols, key=lambda m: m.score).score
    major_sols = sorted(
        [
            m
            for m in major_sols
            if m.score - min_major_score - profile.gap < SOLUTION_PRECISION
        ],
        key=lambda m: (int(1000 * m.score), m._solution_nice()),
    )

    if multiple_warn_level >= 2 and len(major_sols) > 1:
        log.warn("WARNING: multiple major solutions found!")
    log.info(f"Potential major {gene.name} star-alleles for {sample.name}:")
    for i, major_sol in enumerate(major_sols):
        conf = 100 * (min_major_score + SLACK) / (major_sol.score + SLACK)
        log.info(f"  {i + 1:2}: {major_sol._solution_nice()} (confidence: {conf:.0f}%)")
    log.debug("*" * 80)

    minor_sols = []
    for m in minor.estimate_minor(
        gene,
        sample.coverage,
        major_sols,
        solver,
        max_solutions=profile.max_minor_solutions,
    ):
        n = solutions.MinorSolution(
            m.score
            * ((m.major_solution.cn_solution.score + SLACK) / (min_cn_score + SLACK)),
            m.solution,
            m.major_solution,
            profile=profile,
        )
        n.set_diplotype(m.get_diplotype())
        minor_sols.append(n)

    if len(minor_sols) == 0:
        if is_simple:
            print(file=output_file)
        raise AldyException(
            "Aldy could not phase any major solution.\n"
            + "Possible solutions:\n"
            + " - Check the coverage. Extremely low coverage prevents Aldy from "
            + "calling star-alleles.\n"
            + " - Run with --debug parameter and notify the authors of Aldy."
        )
    min_minor_score = min(minor_sols, key=lambda m: m.score).score
    minor_sols = sorted(
        [
            m
            for m in minor_sols
            if m.score - min_minor_score - profile.gap < SOLUTION_PRECISION
        ],
        key=lambda m: (int(1000 * m.score), m._solution_nice()),
    )
    log.debug("*" * 80)

    if multiple_warn_level >= 1 and len(minor_sols) > 1:
        log.warn("WARNING: multiple optimal solutions found!")
    log.info(
        f"{{}} {gene.name} star-alleles for {sample.name}:",
        "Best" if len(minor_sols) == 1 else "Potential",
    )
    if is_aldy:
        print("#" + "\t".join(OUTPUT_COLS), file=output_file)
    simple = [sample.name, gene.name]
    for i, minor_sol in enumerate(minor_sols):
        conf = 100 * (min_minor_score + SLACK) / (minor_sol.score + SLACK)
        log.info(
            f"  {i + 1:2}: {minor_sol.get_major_diplotype()}"
            + f" (confidence={conf:.0f}%)"
        )
        log.info(f"      Minor alleles: {minor_sol._solution_nice()}")
        simple += [
            minor_sol.get_major_diplotype().replace(" ", ""),
            minor_sol.get_minor_diplotype(legacy=True).replace(" ", ""),
        ]
        if is_aldy:
            print(f"#Solution {i + 1}: {minor_sol._solution_nice()}", file=output_file)
            diplotype.write_decomposition(
                sample.name, gene, sample.coverage, i + 1, minor_sol, output_file
            )
        elif is_simple:
            print(simple[-2], simple[-1], sep="\t", end="\t", file=output_file)
    if is_vcf:
        diplotype.write_vcf(sample.name, gene, sample.coverage, minor_sols, output_file)
    elif is_simple:
        print(file=output_file)
    log.debug("[simple]\t{}", "\t".join(simple))

    if report:
        log.info(colorize(f"{gene.name} results:"))
        reported: Set[str] = set()
        for r in minor_sols:
            st = f"  - {r.get_major_diplotype()}"
            if st not in reported:
                log.info(colorize(st))
                log.info(f"    Minor: {r.get_minor_diplotype()}")
                log.info(f"    Legacy notation: {r.get_minor_diplotype(legacy=True)}")
                reported.add(st)
                # Output the activity
                for m in r.solution:
                    mn = [
                        ma.minors[m.minor]
                        for ma in gene.alleles.values()
                        if m.minor in ma.minors
                    ]
                    msg = f"Estimated activity for {m.major_repr()}"
                    if len(mn) == 1 and mn[0].activity:
                        msg = f"{msg}: {mn[0].activity}"
                        if mn[0].evidence:
                            msg = f"{msg} (evidence: {mn[0].evidence})"
                        if mn[0].pharmvar:
                            msg = f"{msg}; see {mn[0].pharmvar} for details"
                    else:
                        msg = f"{msg}: unknown"
                    log.info(f"    {msg}")
                # Calculate coverage of each mutation in a solution
                if multiple_warn_level >= 2:
                    for m, mc, sc in r.get_mutation_coverages(sample.coverage):
                        if abs(mc - sc) > 0.5:
                            log.warn(
                                "    Warning: Coverage of {} is not in line with "
                                "the prediction (predicted: {:.1f}, observed: {:.1f})",
                                gene.get_rsid(m),
                                mc,
                                sc,
                            )

    if (
        any(len(m.major_solution.added) > 0 for m in minor_sols)
        and not sample.is_long_read
    ):
        novels = {
            gene.get_rsid(mm) for m in minor_sols for mm in m.major_solution.added
        }
        log.warn(
            "WARNING: mutations {} suggest presence of a novel major star-allele.\n"
            "However, exact allele contents cannot be determined without the "
            "long-read phasing data. The reported assignments of novel mutations "
            "are for that reason currently randomly assigned.",
            ", ".join(sorted(novels)),
        )

    log.debug("[genotype] gene={}; took={}", gene_db, time.time() - t1)
    return {gene_db: minor_sols}



# 786
# Aldy source: query.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import sys
import textwrap
from natsort import natsorted

from .common import AldyException, log
from .gene import Gene


def query(gene: Gene, query: str):
    """Query a gene instance."""

    minors = {m: a for a, al in gene.alleles.items() for m in al.minors}
    alts = {
        mal.alt_name: (a, m)
        for a, al in gene.alleles.items()
        for m, mal in al.minors.items()
        if mal.alt_name
    }
    if query:
        if query.lower() in map(str.lower, gene.cn_configs):
            query = next(q for q in gene.cn_configs if query.lower() == q.lower())
            print_cn(gene, query)
        elif query.lower() in map(str.lower, gene.alleles):
            query = next(q for q in gene.alleles if query.lower() == q.lower())
            print_majors(gene, query)
        elif query.lower() in map(str.lower, minors):
            query = next(q for q in minors if query.lower() == q.lower())
            print_minors(gene, minors[query], query)
        elif query.lower() in map(str.lower, alts):
            query = next(q for q in alts if query.lower() == q.lower())
            print_minors(gene, *alts[query])
        else:
            log.error(f"ERROR: Cannot parse query '{query}'")
            sys.exit(1)
        return

    name = [f"Gene {gene.name}"]
    if gene.ensembl:
        name.append(f"ENSEMBL: {gene.ensembl}")
    if gene.pharmvar:
        name.append(f"PharmVar ID: {gene.pharmvar}")
    log.info("-" * 80)
    log.info("{}{}", name[0], "" if len(name) == 1 else f" ({', '.join(name[1:])})")
    log.info("-" * 80)

    st, ed = gene.ref_to_chr[0], gene.ref_to_chr[len(gene.seq) - 1]
    log.info(f"{gene.genome} genome locus: {gene.chr}:{st}-{ed} ({gene.refseq})")
    log.info("Strand: " + "3' (reverse)" if gene.strand < 0 else "5'")
    pseudo = ", ".join(f"{p} (ID {i + 1})" for i, p in enumerate(gene.pseudogenes))
    log.info(f"Pseudogenes: {pseudo if pseudo else 'none'}")
    amino = "\n  ".join(textwrap.wrap(gene.aminoacid, width=78))
    log.info(f"Aminoacid:\n  {amino}")

    log.info("Regions of interest:")
    s = [f"  {'Region:':>10}  {gene.name:>25}"]
    if gene.pseudogenes:
        s[-1] += f" {gene.pseudogenes[0]:>25}"
    for r in gene.regions[0]:
        lg = str(gene.regions[0].get(r, "-"))
        lp = str((gene.regions[1] if len(gene.regions) > 1 else {}).get(r, ""))
        s.append(f"  {r:>10}: {lg:>25} {lp:>25}")
    log.info("\n".join(s))

    log.info("Structural alleles (deletions, conservations and fusions):")
    for c, cn in gene.cn_configs.items():
        log.info(f"  {'*'+c+':':>8} {cn.description}")

    log.info("Major star-alleles:")
    for a, allele in natsorted(gene.alleles.items()):
        if "#" in a:  # avoid fusions here
            continue
        log.info(f"  *{a}:")
        m = ",\n                   ".join(
            _print_mutation(gene, m) for m in sorted(allele.func_muts)
        )
        log.info(f"    Key mutations: {m if m else 'none'}")

        mins = ",\n                   ".join(
            f"*{a}" + (f" (*{mi.alt_name})" if mi.alt_name else "")
            for a, mi in natsorted(allele.minors.items())
        )
        log.info(f"    Minor alleles: {mins if mins else 'none'}")


def print_cn(gene: Gene, major: str):
    """Pretty-print a copy number configuration."""
    if major not in gene.cn_configs:
        raise AldyException(f"Structural allele {major} not found in {gene.name}")

    config = gene.cn_configs[major]
    log.info(f"Gene {gene.name}, structural allele {gene.name}*{major}:")

    lg = " ".join(f"{i:<8}" for i in [gene.name] + gene.pseudogenes)
    log.info(f"  Structure: {lg}".rstrip())
    for r in gene.regions[0]:
        lg = " ".join(f"{config.cn[i][r]:<8}" for i, _ in enumerate(gene.regions))
        log.info(f"      {r:>5}: {lg}".rstrip())

    if major in gene.alleles:
        print_majors(gene, major)


def print_majors(gene: Gene, major: str):
    """Pretty-print a major star-allele."""

    if major not in gene.alleles:
        raise AldyException(f"Major star-allele {major} not found in {gene.name}")

    allele = gene.alleles[major]
    log.info(f"Gene {gene.name}, major star-allele {gene.name}*{allele.name}:")
    log.info(f"  Structure: {allele.cn_config}")

    activities = {m.activity for m in allele.minors.values() if m.activity}
    if len(activities) == 1:
        log.info(f"  Activity: {activities.pop()}")
    elif len(activities) > 1:
        log.info(f"  Activity: {' or '.join(activities)} (please check minor alleles)")

    ms = ",\n                 ".join(
        _print_mutation(gene, m) for m in sorted(allele.func_muts)
    )
    log.info(f"  Key mutations: {ms if ms else 'none'}")

    log.info("  Minor star-alleles:")
    for a, minor in natsorted(allele.minors.items()):
        log.info(f"    *{a}:")
        m = ",\n                        ".join(
            _print_mutation(gene, m) for m in sorted(minor.neutral_muts)
        )
        if minor.alt_name:
            log.info(f"      Legacy name: *{minor.alt_name}")
        log.info(f"      Silent mutations: {m if m else 'none'}")


def print_minors(gene: Gene, major: str, minor: str):
    """Pretty-print a minor star-allele."""

    allele = gene.alleles[major]
    if minor not in allele.minors:
        raise AldyException(f"Minor star-allele {minor} not found in {gene.name}")

    log.info(f"\nGene {gene.name}, minor star-allele {gene.name}*{minor}:")
    log.info(f"  Structure: {allele.cn_config}")
    log.info(f"  Major star-allele: {gene.name}*{allele.name}")
    mn = allele.minors[minor]
    if mn.activity:
        log.info(f"  Activity: {mn.activity}")
    if mn.evidence:
        evidence = {"L": "limited", "D": "definitive"}.get(mn.evidence, mn.evidence)
        log.info(f"  Evidence: {evidence}")
    if mn.pharmvar:
        log.info(f"  PharmVar details: {mn.pharmvar}")
    if mn.alt_name:
        log.info(f"  Legacy name: *{mn.alt_name}")

    ms = ",\n                 ".join(_print_mutation(gene, m) for m in allele.func_muts)
    log.info(f"  Key mutations: {ms if ms else 'none'}")

    m = ",\n                    ".join(
        _print_mutation(gene, m) for m in sorted(mn.neutral_muts)
    )
    log.info(f"  Silent mutations: {m if m else 'none'}")


def _print_mutation(gene: Gene, m):
    """Pretty-print a mutation."""

    fields = [
        gene.get_rsid(m, default=False),
        gene.get_refseq(m, from_atg=True),
        f"{gene.refseq}:" + gene.get_refseq(m),
    ]
    if m in gene.mutations and gene.mutations[m][0]:
        fields.append(gene.mutations[m][0])  # type: ignore
    if fields[0] == "-":
        fields = fields[1:]
    return f"{gene.chr}:{m} ({', '.join(fields)})"



# 786
# Aldy source: major.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Dict, Tuple, Any, Optional
from natsort import natsorted
import collections
import copy

from . import lpinterface
from .common import log, json, sorted_tuple
from .gene import MajorAllele, Mutation, Gene
from .coverage import Coverage
from .solutions import CNSolution, MajorSolution, SolvedAllele


def estimate_major(
    gene: Gene,
    coverage: Coverage,
    cn_solution: CNSolution,
    solver: str,
    identifier: int = 0,
    debug: Optional[str] = None,
) -> List[MajorSolution]:
    """
    Estimate optimal major star-alleles.

    :param gene: Gene instance.
    :param profile: Profile instance.
    :param coverage: Read coverage instance.
    :param cn_solution: Copy-number solution for major star-allele calling.
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param identifier: Unique solution identifier. Used for debug purposes.
        Default: 0.
    :param debug: When set, create a `{debug}.major{identifier}.lp` model description
        file for debug purposes.
        Default: `None` (no debug dumps).
    """

    log.debug("*" * 80)
    log.debug("[major] struct= {}", cn_solution._solution_nice())

    alleles, coverage = _filter_alleles(gene, coverage, cn_solution)
    coverage.dump(log.trace)

    # Check if some CN solution has no matching allele
    if set(cn_solution.solution) - set(a.cn_config for a in alleles.values()):
        results: List[MajorSolution] = []
    else:
        results = solve_major_model(
            gene, coverage, cn_solution, alleles, solver, identifier, debug
        )
    # TODO: Check for novel functional mutations and do something with them

    return results


def solve_major_model(
    gene: Gene,
    coverage: Coverage,
    cn_solution: CNSolution,
    allele_dict: Dict[str, MajorAllele],
    solver: str,
    identifier: int = 0,
    debug: Optional[str] = None,
) -> List[MajorSolution]:
    """
    Solves the major star-allele detection problem via integer linear programming.

    :param gene: Gene instance.
    :param coverage: Read coverage instance.
    :param cn_solution: Copy-number solution for major star-allele calling.
    :param allele_dict: Candidate major star-alleles.
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param identifier: Unique solution identifier. Used for debug purposes.
        Default: 0.
    :param debug: When set, create a `{debug}.major{identifier}.lp` model description
        file for debug purposes.
        Default: `None` (no debug dumps).

    .. note::
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Major star-allele identification) for the model explanation.
    """

    model = lpinterface.model("AldyMajor", solver)
    debug_info = json[gene.name]["major"][len(json[gene.name]["major"])]

    # Get the list of _all_ functional mutations present in the sample
    # and the database (intersection)
    func_muts = {
        Mutation(*m)
        for m in gene.mutations
        if gene.is_functional(m) and coverage[Mutation(*m)] > 0
    }
    _print_candidates(gene, allele_dict, coverage, cn_solution, func_muts)

    a: Any = 0

    # Create a binary variable for every possible allele copy
    alleles = {(a, int(0)): allele_dict[a] for a in allele_dict}
    for (an, _), a in list(alleles.items()):
        max_cn = cn_solution.solution[a.cn_config]
        for i in range(1, max_cn):
            alleles[an, i] = alleles[an, 0]
    VA = {a: model.addVar(vtype="B", name=f"A_{a[0]}_{a[1]}") for a in alleles}

    # Make sure that VA[i+1] <= VA[i] (to avoid equivalent solutions)
    for a, ai in alleles.keys():
        if ai > 0:
            model.addConstr(VA[a, ai] <= VA[a, ai - 1], name=f"CORD_{a}_{ai}")

    # Add an error variable for every mutation
    VERR = {
        m: model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{m.pos}_{m.op}")
        for m in func_muts
    }
    constraints = {e: 0 for e in VERR}

    # Add a binary variable for each mutation copy
    # For each present functional mutation, add a binary variable VNEW s.t.
    #   VNEW[m] = 0 <=> sum(VA[a] if m in a) >= 1
    VNEW = {m: model.addVar(vtype="B", name=f"N_{m}") for m in func_muts}

    # Populate the constraints
    for a in alleles:
        for m in alleles[a].func_muts:
            constraints[m] += VA[a]
    for m, v in VNEW.items():
        constraints[m] += v

    # Populate constraints of non-variations (i.e. matches with the reference genome)
    for pos in set(m.pos for m in constraints):
        ref_m = Mutation(pos, "_")  # type: ignore
        constraints[ref_m] = 0
        VERR[ref_m] = model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{pos}_REF")
        for a in alleles:
            if not gene.has_coverage(a[0], pos):
                continue
            # An insertion still contributes to the total coverage at this loci
            if any(ma[0] == pos and ma[1][:3] != "ins" for ma in alleles[a].func_muts):
                continue
            constraints[ref_m] += VA[a]

        # We ensure that only one additional mutation can be selected here (HACK)
        z = model.quicksum(
            v for m, v in VNEW.items() if m[0] == pos and m[1][:3] != "ins"
        )
        model.addConstr(z <= 1, name=f"CONE_{pos}")

    # Each allele must express all of its functional mutations
    debug_info["id"] = identifier
    debug_info["cn"] = str(dict(cn_solution.solution))
    debug_info["data"] = []
    for m, expr in sorted(constraints.items()):
        if coverage.single_copy(m.pos, cn_solution) == 0:
            cov = 0.0
        else:
            cov = coverage[m] / coverage.single_copy(m, cn_solution)
        model.addConstr(expr + VERR[m] <= cov, name=f"CFUNC_{m.pos}_{m.op}")
        model.addConstr(expr + VERR[m] >= cov, name=f"CFUNC_{m.pos}_{m.op}")
        debug_info["data"].append((m[0], m[1], cov))

    # Each CN configuration must be satisfied by corresponding alleles
    for cnf, cnt in cn_solution.solution.items():
        expr = sum(VA[a] for a in VA if alleles[a].cn_config == cnf)
        model.addConstr(expr <= cnt, name=f"CSAT_{cnf}")
        model.addConstr(expr >= cnt, name=f"CSAT_{cnf}")

    # Each functional mutation must be either chosen by some allele or marked as novel:
    #   1 == (m chosen by allele) XOR (m novel) == OR(VA[a] if m in a) XOR VNEW[m]
    for m in func_muts:
        VOR = model.addVar(vtype="B", name=f"OR_{m}")
        m_all = {a for a in alleles if m in alleles[a].func_muts}
        model.addConstr(VOR <= model.quicksum(VA[a] for a in m_all), name="COR")
        for a in m_all:
            model.addConstr(VOR >= VA[a], name="COR")
        VXOR = model.addVar(vtype="B", name=f"XOR_{m}")
        model.addConstr(VXOR <= VNEW[m] + VOR, name="CXOR")
        model.addConstr(VXOR <= 2 - VNEW[m] - VOR, name="CXOR")
        model.addConstr(VXOR >= VNEW[m] - VOR, name="CXOR")
        model.addConstr(VXOR >= VOR - VNEW[m], name="CXOR")
        model.addConstr(VXOR >= 1, name="CXOR")

    # Objective: minimize the absolute sum of errors and the number of novel mutations
    objective = 0
    objective += model.abssum(e for e in VERR.values())

    z = model.addVar(vtype="B", name="NOVEL")
    for m in VNEW:
        model.addConstr(z >= VNEW[m], name=f"NOVEL_UB_{VNEW[m]}")
    model.addConstr(z <= model.quicksum(VNEW[m] for m in VNEW), name="NOVEL_LB")
    objective += coverage.profile.major_novel * z
    objective += 0.1 * model.quicksum(VNEW[m] for m in VNEW)
    model.setObjective(objective)
    if debug:
        model.dump(f"{debug}.{gene.name}.major{identifier}.lp")

    # Solve the model
    lookup = {
        **{model.varName(v): a for a, v in VA.items()},
        **{model.varName(v): m for m, v in VNEW.items()},
    }
    result: Dict[Any, MajorSolution] = {}
    debug_info["sol"] = []
    for status, opt, sol in model.solutions(coverage.profile.gap):
        solved_alleles = sorted_tuple(
            [lookup[s][0] for s in sol if s in lookup and s.startswith("A_")]
        )
        novel_muts = sorted_tuple(
            [lookup[s] for s in sol if s in lookup and s.startswith("N_")]
        )
        if (solved_alleles, novel_muts) not in result:
            solution = collections.Counter(
                SolvedAllele(gene, major=a) for a in solved_alleles
            )
            debug_info["sol"].append(
                dict(collections.Counter(a for a in solved_alleles))
            )
            sol = MajorSolution(
                score=opt,
                solution=solution,
                cn_solution=cn_solution,
                added=list(novel_muts),
            )
            log.debug(
                f"[major] status= {status}; opt= {opt:.2f}; "
                + f"solution= {sol._solution_nice()}"
            )
            result[solved_alleles, novel_muts] = sol
    if not result:
        log.debug("[major] solution= []")
    return list(result.values())


def _filter_alleles(
    gene: Gene, coverage: Coverage, cn_solution: CNSolution
) -> Tuple[Dict[str, MajorAllele], Coverage]:
    """
    Filter out low-quality mutations and alleles that are not expressed.

    :returns: Expressed alleles and the coverage of high-quality variants.
    """

    def filter_fns(cov, mut):
        cond = cov.basic_filter(mut, cn=coverage.profile.cn_max)
        if mut.op != "_":
            cond = cond and cov.basic_filter(
                mut, cn=cn_solution.position_cn(mut.pos) + 0.5
            )
        return cond

    cov = coverage.filtered(Coverage.quality_filter)
    cov = cov.filtered(filter_fns)

    alleles = copy.deepcopy(gene.alleles)
    for an, a in natsorted(gene.alleles.items()):
        if a.cn_config not in cn_solution.solution:
            del alleles[an]
        elif any(cov[m] <= 0 for m in a.func_muts):
            s = (f"{gene.get_rsid(m)}" for m in a.func_muts if cov[m] <= 0)
            log.trace("[major] removing {} because of {}", an, " and ".join(s))
            del alleles[an]

    return alleles, cov


def _print_candidates(
    gene,
    alleles: Dict[str, Any],
    coverage: Coverage,
    cn_solution: CNSolution,
    muts: set,
):
    """Pretty-prints the list of allele candidates and their functional mutations."""

    def print_mut(m):
        copies = (
            coverage[m] / (coverage.total(m) / cn_solution.position_cn(m.pos))
            if cn_solution.position_cn(m.pos) and coverage.total(m)
            else 0
        )
        g = gene.region_at(m.pos)
        return (
            f"  {gene.get_rsid(m):12} {str(m):15} "
            + f"{gene.get_refseq(m, from_atg=True):10} "
            + f"(cov={coverage[m]:4}, cn= {copies:3.1f}; "
            + f"region={g[1] if g else '?'}; "
            + f"impact={gene.get_functional(m)}; "
            + ")"
        )

    log.debug("[major] candidate mutations=")
    muts = set(m for a in alleles for m in alleles[a].func_muts) | set(muts)
    for m in sorted(muts):
        log.debug(print_mut(m))
        als = natsorted(
            f"*{a:8}"
            for a, al in gene.alleles.items()
            if m in al.func_muts and "#" not in a
        )
        log.debug(
            "    {}",
            "\n    ".join(" ".join(als[i : i + 6]) for i in range(0, len(als), 6)),
        )
    log.debug("[major] candidate alleles=")
    muts = muts.copy()
    for a in natsorted(alleles):
        log.debug("  *{} (struct= *{})", a, alleles[a].cn_config)
        for m in sorted(alleles[a].func_muts):
            if m in muts:
                muts.remove(m)
            log.debug("  {}", print_mut(m))
    if len(muts) > 0:
        log.debug("  Other mutations:")
        for m in sorted(muts):
            a = (f"*{a}" for a, b in gene.alleles.items() if m in b.func_muts)
            log.debug("  {}, alleles={})", print_mut(m)[:-1], ", ".join(a))



# cython: embedsignature=True
from .variant cimport Variant

cdef class FailedContig:
    cdef public bint target_not_found
    cdef public bint is_low_quality
    cdef public bint failed_anyway

cdef class Contig:
    cdef Variant target
    cdef list pileup, targetpileup
    cdef public bint qc_passed, failed
    cdef int donwsample_lim
    cdef object lt_genomic_index, rt_genomic_index
    cdef int start, end
    cdef str lt_reference_seq, rt_reference_seq
    cdef str lt_target_block_reference_seq, rt_target_block_reference_seq
    cdef bint is_non_complex_at_target_pos
    cdef str target_ref, target_alt
    cdef double low_consensus_thresh
    cdef public int is_target_right_aligned
    cdef public object contig_dict
    cdef public str lt_consensus_seq, rt_consensus_seq, indel_seq
    cdef public str lt_target_block_consensus_seq, rt_target_block_consensus_seq
    cdef public list lt_consensus_scores, rt_consensus_scores, mismatches, non_target_indels, gaps
    cdef public list lt_target_block_consensus_scores, rt_target_block_consensus_scores
    cdef public int mapq, lt_end_pos
    cdef public double low_qual_mapping_rate
    cdef public dict qc_stats
    cdef public tuple splice_pattern



# 786
# Aldy source: cn.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, List, Tuple, Optional
from math import ceil
from natsort import natsorted
from functools import partial
import copy
import re

from . import lpinterface
from .common import log, json, sorted_tuple, AldyException
from .gene import CNConfig, CNConfigType, Gene
from .coverage import Coverage
from .solutions import CNSolution
from .profile import Profile


def estimate_cn(
    gene: Gene,
    profile: Profile,
    coverage: Optional[Coverage],
    solver: str,
    debug: Optional[str] = None,
) -> List[CNSolution]:
    """
    Estimate the optimal copy number configuration for a sample given a gene and
    coverage information.

    :param gene: Gene instance.
    :param profile: Profile instance.
    :param coverage: Read coverage instance.
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param debug: When set, create a `{debug}.cn.lp` model description file for
        debug purposes.
        Default: `None` (no debug dumps).

    :returns: List of optimal copy number configurations.
    """

    log.debug("\n" + "*" * 80)

    if profile.cn_solution:
        return [_parse_user_solution(gene, profile.cn_solution)]
    elif not gene.do_copy_number:
        basic = [
            k for k, v in gene.cn_configs.items() if v.kind == CNConfigType.DEFAULT
        ]
        cn = 2
        if profile.male and gene.chr in ["X", "Y"]:
            cn = 1
        return [_parse_user_solution(gene, [basic[0]] * cn)]
    else:
        assert coverage, "Coverage not provided"

        # TODO: filter CN configs with non-present alleles
        max_observed_cn = 1 + max(
            ceil(coverage.region_coverage(gi, r))
            for gi, g in enumerate(gene.regions)
            for r in g
        )
        region_cov = {
            r: (
                coverage.region_coverage(0, r),
                coverage.region_coverage(1, r) if len(gene.regions) > 1 else 0.0,
            )
            for r in gene.unique_regions
        }
        total_cov = sum(r0 + r1 for r0, r1 in region_cov.values())
        min_cov = min(
            sum(sum(v.values()) for v in gene.cn_configs[c].cn) for c in gene.cn_configs
        )
        if total_cov < min_cov / 2.0:
            raise AldyException(
                f"Coverage for {gene.name} too low for copy number calling."
            )

        configs = _filter_configs(gene, coverage)
        log.debug("[cn] candidates= {}", ", ".join(natsorted(configs)))
        log.debug("[cn] max_cn= {}", max_observed_cn)
        _print_coverage(gene, coverage)

        fusion_support = None
        if coverage.sam._fusion_counter:
            fusion_support = {
                fn: (a / b) if b else 0.0
                for fn, [a, b] in coverage.sam._fusion_counter.items()
            }
        sol = solve_cn_model(
            gene,
            profile,
            configs,
            max_observed_cn,
            region_cov,
            solver,
            debug,
            fusion_support,
        )

        return sol


def solve_cn_model(
    gene: Gene,
    profile: Profile,
    cn_configs: Dict[str, CNConfig],
    max_cn: int,
    region_coverage: Dict[str, Tuple[float, float]],
    solver: str,
    debug: Optional[str] = None,
    fusion_support: Optional[Dict[str, float]] = None,
) -> List[CNSolution]:
    """
    Solve the copy number estimation problem (instance of the closest vector problem).

    :param gene: Gene instance.
    :param profile: Profile instance.
    :param cn_configs: Available copy number configurations (vectors).
    :param max_cn: Maximum allowed copy number.
    :param region_coverage: Observed copy number of the main gene and the pseudogene
        for each genic region.
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param gap: Optimality gap. Non-zero values enable non-optimal solutions.
        Default: 0 (only optimal solutions).
    :param debug: When set, create a `{debug}.cn.lp` model description file for
        debug purposes.
        Default: `None` (no debug dumps).
    :param fusion_support: Dictionary that contains read support of each available
        fusion. Used only for the long-read fusion calling.
        Default is `None` (all fusions are treated equally).
    :returns: List of optimal copy-number solutions.

    .. note::
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Copy number and structural variation estimation)
        for the detailed description of ILP model.
    """

    model = lpinterface.model("AldyCN", solver)
    debug_info = json[gene.name]["cn"]

    del_allele = gene.deletion_allele()

    # List of CN configurations (a.k.a. structures). Each configuration is a binary
    # variable. Each structure is defined by `('structure_name', number)`.
    # When `number` ∈ {0, -1}, the configuration is complete with *all* pseudogenes
    # included (a diploid genome must contain exactly *2* complete configurations).
    # When `number` > 0, the configuration describes only the main gene and not the
    # pseudogene (there can be many such configurations).
    # N.B. (3/2022) this step filters "weak" fusions without long-read support
    structures: Dict[Tuple[str, int], CNConfig] = {
        (name, 0): structure
        for name, structure in cn_configs.items()
        if not fusion_support
        or name == "1"
        or (del_allele and name == del_allele)
        or (name in fusion_support and fusion_support[name] >= 1 / (2 * max_cn))
    }
    for a, ai in list(structures.keys()):
        structures[a, -1] = copy.deepcopy(structures[a, 0])
        if cn_configs[a].kind != CNConfigType.DEFAULT:
            continue
        for i in range(1, max_cn):
            structures[a, i] = copy.deepcopy(structures[a, 0])
            for g in range(1, len(structures[a, i].cn)):
                # If this is a pseudogene, remove a copy (create a "weak configuration")
                structures[a, i].cn[g] = {
                    r: v - 1 for r, v in structures[a, i].cn[g].items()
                }
    # Add "fake" pseudogenes to handle pseudogene CN changes or CN noise
    if len(gene.regions) > 1 and del_allele:
        for i in range(max_cn):
            structures["PSEUDO", i + 1] = CNConfig(
                copy.deepcopy(structures[del_allele, 0].cn),
                CNConfigType.DELETION,
                set(),
                "pseudogene",
            )

    # Add a binary variable for each CN structure
    VCN = {
        (a, ai): model.addVar(vtype="B", name=f"CN_{a}_{ai}") for a, ai in structures
    }

    # We assume diploid genome, so the number of the complete configurations must be 2
    diplo_inducing = model.quicksum(VCN[a] for a in VCN if a[1] <= 0)
    model.addConstr(diplo_inducing <= 2, name="CDIPLO")
    model.addConstr(diplo_inducing >= 2, name="CDIPLO")

    # Ensure that we cannot link any allele to the whole-gene deletion
    if del_allele:
        for (a, ai), v in VCN.items():
            if a != del_allele:
                model.addConstr(v + VCN[del_allele, -1] <= 1, name=f"CDEL_{a}_{ai}")

    # Ensure that binary transformation is properly formed (i.e. A_i <= A_{i-1})
    for a, ai in structures:
        # The second haplotype (-1) is only present if the first one (0) is there
        if ai == -1:
            model.addConstr(VCN[a, ai] <= VCN[a, 0], name=f"CORD_{a}_{ai}")
        # Ignore 1, because A[1] = 1 && A[0] = 0 is valid (e.g. *13/*13+*1)
        elif ai > 1:
            model.addConstr(VCN[a, ai] <= VCN[a, ai - 1], name=f"CORD_{a}_{ai}")

    # Add error variables
    VERR, VERR_GENE = {}, {}
    for r, (exp_cov0, exp_cov1) in region_coverage.items():
        debug_info["data"][r] = (exp_cov0, exp_cov1)
        expr, expr_gene = 0, 0
        for s, structure in structures.items():
            if r in structure.cn[0]:
                expr += structure.cn[0][r] * VCN[s]
                expr_gene += structure.cn[0][r] * VCN[s]
            if len(structure.cn) > 1 and r in structure.cn[1]:
                expr -= structure.cn[1][r] * VCN[s]

        if r not in gene.unique_regions:
            continue

        VERR_GENE[r] = model.addVar(
            name=f"EG_{r}", lb=-profile.cn_max, ub=profile.cn_max
        )
        model.addConstr(expr_gene + VERR_GENE[r] <= exp_cov0, name=f"CG_COV_{r}")
        model.addConstr(expr_gene + VERR_GENE[r] >= exp_cov0, name=f"CG_COV_{r}")

        scale = max(exp_cov0, exp_cov1) + 1
        VERR[r] = model.addVar(name=f"E_{r}", lb=-profile.cn_max, ub=profile.cn_max)
        model.addConstr(
            expr / scale + VERR[r] <= (exp_cov0 - exp_cov1) / scale, name=f"C_COV_{r}"
        )
        model.addConstr(
            expr / scale + VERR[r] >= (exp_cov0 - exp_cov1) / scale, name=f"C_COV_{r}"
        )

    # Objective component 1: minimize the sum of absolute errors.
    # PCE_REGION (in CYP2D7) is penalized with an extra score (important fusion marker).
    DIFF_COEFF = profile.cn_diff / len(gene.unique_regions)
    o_diff = DIFF_COEFF * model.abssum(
        VERR.values(), coeffs={"E_pce": profile.cn_pce_penalty}
    )

    # Objective component 2: minimize the main gene fitness.
    FIT_COEFF = profile.cn_fit / len(gene.unique_regions)
    o_fit = FIT_COEFF * model.abssum(VERR_GENE.values())

    # Objective component 3: minimize the total number of present alleles
    #                        (maximum parsimony)
    PARSIMONY_PENALTY = 10.0 / len(gene.unique_regions)
    PARSIMONY_PENALTY *= 0.75
    penalty = {s: PARSIMONY_PENALTY for s, _ in VCN}
    # Also penalize the left fusions as they are less likely
    for n, sv in gene.cn_configs.items():
        if n in penalty and sv.kind == CNConfigType.RIGHT_FUSION:
            penalty[n] += PARSIMONY_PENALTY * profile.cn_fusion_right
        if n in penalty and sv.kind == CNConfigType.LEFT_FUSION:
            penalty[n] += PARSIMONY_PENALTY * profile.cn_fusion_left
    o_pars = profile.cn_parsimony * model.quicksum(
        penalty[s] * v for (s, _), v in VCN.items()
    )

    model.setObjective(o_diff + o_fit + o_pars)
    if debug:
        model.dump(f"{debug}.{gene.name}.cn.lp")

    # Solve the model
    lookup = {model.varName(v): a for (a, _), v in VCN.items()}
    result: dict = {}
    for status, opt, sol in model.solutions(profile.gap):
        sol_tuple = sorted_tuple(
            lookup[v] for v in sol if lookup[v] not in [del_allele, "PSEUDO"]
        )
        if sol_tuple not in result:
            result[sol_tuple] = CNSolution(gene, opt, list(sol_tuple))
            log.debug(
                f"[cn] status= {status}; opt= {opt:.2f} "
                + "(diff= {:.2f}, fit= {:.2f}, pars= {:.2f}) "
                + f"solution= {result[sol_tuple]}",
                model.getValue(o_diff),
                model.getValue(o_fit),
                model.getValue(o_pars),
            )
    if not result:
        log.debug("[cn] solution= []")

    debug_info["sol"] = [dict(r.solution) for r in result.values()]
    return list(result.values())


def _filter_configs(gene: Gene, coverage: Coverage) -> Dict[str, CNConfig]:
    """
    Filter out low-quality mutations and copy number configurations that are not
    supported by the remaining mutations.
    """

    cov = coverage.filtered(
        partial(
            Coverage.basic_filter,
            thres=coverage.profile.threshold / coverage.profile.cn_max,
        )
    )
    configs = copy.deepcopy(gene.cn_configs)
    for an in natsorted(gene.cn_configs):
        if an not in gene.alleles:
            continue  # This is just a CN configuration w/o any mutations
        bad_alleles = []
        for a in gene.cn_configs[an].alleles:
            if any(cov[m] <= 0 for m in gene.alleles[a].func_muts):
                bad_alleles.append(a)
        if len(bad_alleles) == len(gene.cn_configs[an].alleles):
            log.trace(f"[cn] removing *{an} due to low support")
            del configs[an]
    return configs


def _print_coverage(gene: Gene, coverage: Coverage) -> None:
    """Pretty-print the region coverage."""

    log.debug("[cn] coverage=")
    gname_ = re.split(r"(\d.+)", gene.name)
    gname = gname_[1] if len(gname_) > 1 else gene.name
    pname = ""
    if gene.pseudogenes:
        pname_ = re.split(r"(\d.+)", gene.pseudogenes[0])
        pname = pname_[1] if len(pname_) > 1 else gene.pseudogenes[0]
    log.debug(f"  {'':5} {gname:4} {pname:4}")
    for r in gene.regions[0]:
        g = coverage.region_coverage(0, r)
        if pname:
            p = coverage.region_coverage(1, r)
            if r in gene.unique_regions:
                log.debug(f"  {r:5} {g:4.1f} {p:4.1f} = {g - p:4.1f}")
            else:
                log.debug(f"  {r:5} {g:4.1f} {p:4.1f}")
        else:
            log.debug(f"  {r:5}: {g:4.1f}")


def _parse_user_solution(gene: Gene, sols: List[str]) -> CNSolution:
    """
    Parse user-provided copy number solutions.

    :param gene: Gene instance.
    :param sols: List of valid copy number configurations.
    :returns: User-provided copy number solution.
    :raise: :py:class:`aldy.common.AldyException` if a user-provided solution does
        not match the gene database.
    """

    for sol in sols:
        if sol not in gene.cn_configs:
            raise AldyException(
                "The copy number solution contains unknown copy number configuration "
                f"{sol}. Please run 'aldy query {gene.name}' for the list of "
                "the valid configurations"
            )
    s = CNSolution(gene, 0, sols)  # type: ignore
    log.debug("[cn] result= {} (provided)", s)
    return s
